/*
 * Program        : ampis.c
 *
 * Objectives     : Finding a subset of items in order to give a maximum profit in MPIS by MPIS_Alg
 *
 * Assumptions    : Item ID in dataset file should start from zero.
 *                  I.e. if there are NUM items in the dataset, 
 *                  the item ID should be from 0 to NUM-1.
 *
 * Variations     : 
 *
 * Platform       : SunOS/Linux
 * Compiler       : gcc
 *
 * Programmer     : Raymond Chi-Wing Wong
 * Creation Date  : Oct, 2002
 * Last Update    : Jun 9, 2004
 *
 * Reference      :
 *   1. J. Han, J. Pei, and Y. Yin, 
 *      "Mining Frequent Patterns without Candidate Generation", 
 *      Proc. ACM SIGMOD 2000.
 *   2. R. C.-W. Wong, A. W.-C. Fu and K. Wang
 *      "MPIS: Maximal-Profit Item Selection with Cross-Selling Considerations"
 *      The 2003 IEEE International Conference on Data Mining (ICDM)
 *      
 */


#include <stdio.h>
#include <stdlib.h>

// need to define -DM_TIME to measure time
#ifdef M_TIME
#include <sys/resource.h>
#endif

#include "ampis.h"
#include "readFP.h"
#include "fpt.h"
#include "fpt_mpis.h"
#include "qsort.h"


// global variable
int MAX_STR = 50;	// max. length of the string (used in the name of the file)
int maxTransSize = 90;	// max. transaction size

int NO_SLOT = 7989917;		// no. of slots used in the hash table
int H_START = 26127;		// start index used in the hash table

int noOfTrans = 9832;	// no. of transactions (dummy; will be replaced by the config file)
int noOfItem = 1000;	// no. of items (dummy; will be replaced by the config file)
int J = 500;			// no. of items to be selected (dummy; will be replaced by the config file)

// the following variable is made globally for optimization
int *d;		// a set of unselected elements used 
int noOfD;	// no. of unselected elements

int *sel;	// a set of selected elements
int noOfSel;	// no. of selected elements

int *d2;	// another set of unselected elements used
int noOfD2;	// no. of unselected elements 

int *sel2;	// another set of selected elements
int noOfSel2;	// no. of selected elements

short **matrix;	// a matrix storing the co-occurence
short *infl;	// an array storing the influence of item

// data structure
struct hash_s
{
	int noOfElt;	// no. of hash entries
	short **elt;	// entries of itemsets
	int *freq;	// freq. of the entries
};

typedef struct hash_s hash;
typedef hash *hashPtr;

hashPtr hashMatrix;

// main3 is used for debug
void main3()
{
	int *s;
	short *S;
	//int J = 3;

	//int noOfItem = 6;
	int itemNo = 0;
	int i = 0;

	noOfItem = 6;
	J = 5;
	s = (int *) malloc(sizeof(int)*J);
	S = (short *) malloc(sizeof(int)*noOfItem);

	S[0] = 1;
	S[1] = 2;
	S[2] = 3;
	S[3] = 4;
	S[4] = 5;

	transform2Select(s, S, J, S, itemNo);

	for (i = 0; i < J; i++)
	{
		printf("%d ", s[i]);
	}
	printf("\n");
}

// testSort is used to test the sorting algorithm
int testSort()
{
	int elt[10];
	int orgElt[10];

	int index[10];
	//int *index2;
	short *index3;

	int noOfElt = 10;

	int i;

	int J =7;

	int noOfJ = 0;


	elt[0] = 10;
	elt[1] = 0;
	elt[2] = 0;
	elt[3] = 0;
	elt[4] = 5;
	elt[5] = 0;
	elt[6] = 100;
	elt[7] = 60;
	elt[8] = 0;
	elt[9] = 65;

	for (i = 0; i < noOfElt; i++)
	{
		orgElt[i] = elt[i];
	}

	for (i = 0; i < noOfElt; i++)
	{
		index[i] = i;
	}

	printf("Before sorting...\n");
	for (i = 0; i < noOfElt; i++)
	{
		printf("%d ", elt[i]);
	}
	printf("\n");

	//void quicksort_i(int *sorted_index, int *value, int start_index, int end_index)
	//quicksort_i_indShort(index, elt, 0, noOfElt-1);
	//index2 = sort(elt, noOfElt);
	//short *getBestJItem2(int *element, int noOfElt, int boundJ, int *noOfJ)
	index3 = getBestJItem2(elt, noOfElt, J, &noOfJ);	

	printf("no. of J : %d\n", noOfJ);
	printf("After sorting...\n");
	for (i = 0; i < J; i++)
	{
		printf("%d ", orgElt[index3[i]]);
	}
	printf("\n");

	
	printf("After sorting(index)...\n");
	for (i = 0; i < J; i++)
	{
		printf("%d ", index3[i]);
	}
	printf("\n");

	return 0;
}

// main program
int main()
{
	char *fpConfigFilename = "configSample";
#ifdef M_TIME
 float  userTime, sysTime;
 struct rusage myTime1, myTime2, myTime3;
#endif
#ifdef M_TIME
	struct rusage start_ru, end_ru;
	struct rusage start_ru_m1;
	struct rusage start_ru_m2;
	struct rusage start_ru_m3;
	struct rusage start_ru_m4;
	struct rusage start_ru_m5;
	struct rusage start_ru_m6;
#endif 
	FILE *lhfilePtr;
	FILE *sfilePtr;
	FILE *mfilePtr;
	FILE *afilePtr;
	FILE *ffilePtr;
	FILE *profFilePtr;
	FILE *confFilePtr;
#ifdef M_TIME
	FILE *timeFilePtr;
	int userTimeS, userTimeM;
	int sysTimeS, sysTimeM;
#endif

	transPtr *data;
	int *indCount;
	int *profit;


	int noOfNonZero;

	short **sortEstIndex;

	largeitemPtr *freqitemset, *freqitemset2;
	int upperLimit, upperLimit2;

	double *L, *H;
	int *s;

	int *sortLIndex, *chooseFlag;
	double *tempL;
	double minL;

	double *b;

	int indexB;

	int noOfChoose; // for pruning step
	int *chooseIndex;
	int indexChooseB;

	int noOfSelect;
	//int *selectS;

	int *prof;

	int isRemove;

	double *revNo;

	STRING dataFilename; // = "datawith10000pat.out";
	STRING freqFilename; // = "datawith10000pat.fp";
	STRING twoFilename; // = "datawith10000pat.fp2";

	STRING lhfilename; //= "lh10000.txt";
	STRING sfilename; // = "s10000.txt";
	STRING mfilename; // = "m10000.txt";
	STRING afilename; // = "a10000.txt";
	STRING ffilename; // = "f10000.txt";
	STRING profFilename; // = "profit.txt";
	STRING confFilename = "conf.txt";
#ifdef M_TIME
	STRING timeFilename = "time.txt";
#endif

	int i, j, k;

	double minB;//, maxB;
	
	short **S;
	int *noOfEltS;

	int tempNo;

	int itemAdd, itemRemove;

	double totalProfit, origTotalProfit, naiveTotalProfit;

	// ------------------------------
	// initialization
	// ------------------------------
	dataFilename = (STRING) malloc(sizeof(char)*MAX_STR); // = "datawith10000pat.out";
	freqFilename = (STRING) malloc(sizeof(char)*MAX_STR); ; // = "datawith10000pat.fp";
	twoFilename = (STRING) malloc(sizeof(char)*MAX_STR); ; // = "datawith10000pat.fp2";

	lhfilename = (STRING) malloc(sizeof(char)*MAX_STR); ; //= "lh10000.txt";
	sfilename = (STRING) malloc(sizeof(char)*MAX_STR); ; // = "s10000.txt";
	mfilename = (STRING) malloc(sizeof(char)*MAX_STR); ; // = "m10000.txt";
	afilename = (STRING) malloc(sizeof(char)*MAX_STR); ; // = "a10000.txt";
	ffilename = (STRING) malloc(sizeof(char)*MAX_STR); ; // = "f10000.txt";
	profFilename = (STRING) malloc(sizeof(char)*MAX_STR); ; // = "profit.txt";
	//confFilename = (STRING) malloc(sizeof(char)*MAX_STR); ; // = "conf.txt";

	// ------------------------------
	// configure
	// ------------------------------
	confFilePtr = (FILE *) fopen(confFilename, "r");
	if (confFilePtr == NULL)
	{
		printf("File %s does not exist!\n", confFilename);
		return 0;
	}
	fscanf(confFilePtr, "%d", &J);				//1
	fscanf(confFilePtr, "%d", &noOfTrans);		//2
	fscanf(confFilePtr, "%d", &noOfItem);		//3
	fscanf(confFilePtr, "%s", dataFilename);	//4
	fscanf(confFilePtr, "%s", freqFilename);	//5
	fscanf(confFilePtr, "%s", twoFilename);		//6
	fscanf(confFilePtr, "%s", lhfilename);		//7
	fscanf(confFilePtr, "%s", sfilename);		//8
	fscanf(confFilePtr, "%s", mfilename);		//9
	fscanf(confFilePtr, "%s", afilename);		//10
	fscanf(confFilePtr, "%s", ffilename);		//11
	fscanf(confFilePtr, "%s", profFilename);	//12
	fclose(confFilePtr);

	// ------------------------------
	// print the config info
	// ------------------------------
	printf("J = %d\n", J);				//1
	printf("no. of trans = %d\n", noOfTrans);		//2
	printf("no. of items = %d\n", noOfItem);		//3
	printf("data file name = %s\n", dataFilename);	//4
	printf("freq item file name = %s\n", freqFilename);	//5
	printf("freq item file name with size two = %s\n", twoFilename);		//6
	printf("lh file = %s\n", lhfilename);		//7
	printf("s file = %s\n", sfilename);		//8
	printf("m file = %s\n", mfilename);		//9
	printf("a file = %s\n", afilename);		//10
	printf("f file = %s\n", ffilename);		//11
	printf("profit file = %s\n", profFilename);	//12


	sfilePtr = (FILE *) fopen(sfilename, "w");
	mfilePtr = (FILE *) fopen(mfilename, "w");
	afilePtr = (FILE *) fopen(afilename, "w");
	profFilePtr = (FILE *) fopen(profFilename, "r");
	if (profFilePtr == NULL)
	{
		printf("File %s does not exist!\n", profFilename);
		return 0;
	}
#ifdef M_TIME
	timeFilePtr = (FILE *) fopen(timeFilename, "w");
#endif

	S = (short **) malloc(sizeof(short *)*noOfItem);
	noOfEltS = (int *) malloc(sizeof(int)*noOfItem);

	L = (double *) malloc(sizeof(double)*noOfItem);
	H = (double *) malloc(sizeof(double)*noOfItem);
	s = (int *) malloc(sizeof(int)*J);


	profit = (int *) malloc(sizeof(int)*noOfItem);
	// Profit Initialization
	for (i = 0; i < noOfItem; i++)
	{
		//profit[i] = 1;
		fscanf(profFilePtr, "%d\n", &(profit[i]));
	}

	// ------------------------------
	// read the data set
	// ------------------------------
	printf("Reading data...\n");
	data = readData(dataFilename, noOfTrans);

	
	// ------------------------------
	// Process/Create FP-tree
	// ------------------------------
	//read the fp-tree for FREQUENT ITEMSET
	freqitemset = readFP(freqFilename, &upperLimit);

	// Obtain the support of size two by fp-tree
	freqitemset2 = readFP(twoFilename, &upperLimit2);





	d = (int *) malloc(sizeof(int)*maxTransSize);
	sel = (int *) malloc(sizeof(int)*maxTransSize);

	d2 = (int *) malloc(sizeof(int)*maxTransSize);
	sel2 = (int *) malloc(sizeof(int)*maxTransSize);




 /* read input parameters --------------------------*/
 printf("input\n");
 input(fpConfigFilename);
 expectedK = maxTransSize;
 threshold = 0;
 numItem = noOfItem;
 numTrans = noOfTrans;
 strcpy(dataFile, dataFilename);

#ifdef M_TIME
 getrusage(RUSAGE_SELF,&myTime1);
#endif

 printf("\npass1\n");
 pass1();

 if (numLarge[0] > 0) {

	/* create FP-tree --------------------------*/
 	printf("\nbuildTree\n");
 	buildTree();
	f_init(noOfItem);

	revNo = (double *) malloc(sizeof(double)*noOfItem);

	for (i = 0; i < noOfItem; i++)
	{
		revNo[i] = 1.0/support0[i];
	}

#ifdef M_TIME
	getrusage(RUSAGE_SELF,&myTime2);
#endif

	//testFun1();

 }




#ifdef DEBUG
ffilePtr = (FILE *) fopen(ffilename, "w");
fprintf(ffilePtr, "The Frequent Item set of size 1:\n");
for (i = 0; i < freqitemset2[0]->noOfItemSet; i++)
{
	fprintf(ffilePtr, "%d\t%d\n", freqitemset2[0]->item[i]->itemNo[0], freqitemset2[0]->item[i]->freq);
}
fclose(ffilePtr);
#endif

#ifdef M_TIME
	getrusage(RUSAGE_SELF, &start_ru);
#endif

	//** need to delete other sorting
	// Sorting the freqitemset of size 2
	printf("sorting the freq item set of size 2...\n");
	sortFreqItemSet(freqitemset2);

	// Obtain the matrix of size two
	printf("Obtaingin the matrix of size two...\n");
	hashMatrix = (hashPtr) malloc(sizeof(hash)*NO_SLOT);
	for (i = 0; i < NO_SLOT; i++)
	{
		hashMatrix[i].noOfElt = 0;
	}
	transform2Matrix(freqitemset2, noOfItem);

	
	// free the support of size two of fp-tree
	for (i = 0; i < upperLimit2; i++)
	{
		for (j = 0; j < freqitemset2[i]->noOfItemSet; j++)
		{
			free(freqitemset2[i]->item[j]->itemNo);
			free(freqitemset2[i]->item[j]);
		}
		free(freqitemset2[i]->item);
		free(freqitemset2[i]);
	}
	free(freqitemset2);


#ifdef M_TIME
	getrusage(RUSAGE_SELF, &start_ru_m1);
#endif

	infl = (short *) malloc(sizeof(short)*noOfItem);
	memset(infl, 0, sizeof(short)*noOfItem);


#ifdef M_TIME
	printf("****MODE M_TIME is used\n");
#endif

#ifdef MATRIX
	printf("****MODE MATRIX is used\n");
#endif

#ifdef LARGE_J
	printf("****MODE LARGE_J is used\n");
#endif

#ifdef FP_SCAN
	printf("****MODE FP_SCAN is used\n");
#endif



	// Derive Individual Count
	printf("Deriving individual count...\n");
	indCount = deriveIndividualCount(data, freqitemset, noOfTrans, noOfItem);

	//***RAY (free freqitemset)
	// free the FP-tree for FREQUENT ITEMSET
	for (i = 0; i < upperLimit; i++)
	{
		for (j = 0; j < freqitemset[i]->noOfItemSet; j++)
		{
			free(freqitemset[i]->item[j]->itemNo);
			free(freqitemset[i]->item[j]);
		}
		free(freqitemset[i]->item);
		free(freqitemset[i]);
	}
	free(freqitemset);

#ifdef DEBUG2
	fprintf(sfilePtr, "\n***Individual Count is :\n");
	printf("\n***S***Individual Count is :\n");
	for (i = 0; i < noOfItem; i++)
	{
		fprintf(sfilePtr, "%d\t%d\n", i, indCount[i]);
		printf("***S***%d\t%d\n", i, indCount[i]);
	}
#endif


#ifdef M_TIME
	// Mearue Time
	getrusage(RUSAGE_SELF, &start_ru_m2);

	printf("Main Step 1\n");
	printf("------------\n");
	printf("***T**************%d*******\n", J);
	fprintf(timeFilePtr, "Main Step 1(Individual Count Derivation)\n");
	fprintf(timeFilePtr, "---------------------------------------\n");
	printf("***T***Main Step 1(Individual Count Derivation)\n");
	printf("***T***---------------------------------------\n");

	userTimeS = (int) (start_ru_m2.ru_utime.tv_sec - start_ru_m1.ru_utime.tv_sec);
	userTimeM = (int) (start_ru_m2.ru_utime.tv_usec - start_ru_m1.ru_utime.tv_usec);

	sysTimeS = (int) (start_ru_m2.ru_stime.tv_sec - start_ru_m1.ru_stime.tv_sec);
	sysTimeM = (int) (start_ru_m2.ru_stime.tv_usec - start_ru_m1.ru_stime.tv_usec);

	printf("User time: %ds\n", userTimeS);
	printf("User time: %d micros\n", userTimeM);
	fprintf(timeFilePtr, "User time: %ds\n", userTimeS);
	fprintf(timeFilePtr, "User time: %d micros\n", userTimeM);
	printf("***T***User time: %ds\n", userTimeS);
	printf("***T***User time: %d micros\n", userTimeM);

	printf("System time: %ds\n", sysTimeS);
	printf("System time: %d micros\n", sysTimeM);
	fprintf(timeFilePtr, "System time: %ds\n", sysTimeS);
	fprintf(timeFilePtr, "System time: %d micros\n", sysTimeM);
	printf("***T***System time: %ds\n", sysTimeS);
	printf("***T***System time: %d micros\n", sysTimeM);
#endif 

	// ------------------------------
	// Calculate Estimate Value
	// ------------------------------
	printf("Calculating the estimate Value...\n");


#ifdef DEBUG2
	// Print the estimate value
	fprintf(sfilePtr, "\n***Estimate Value :\n");
	printf("\n***S***Estimate Value :\n");
	estValue = calculateEst(profit, indCount, freqitemset2, noOfItem, matrix);
	for (i = 0; i < noOfItem; i++)
	{
		for (j = 0; j < noOfItem; j++)
		{
			fprintf(sfilePtr, "%d ", estValue[i][j]);
			printf("***S***%d ", estValue[i][j]);
		}
		fprintf(sfilePtr, "\n");
		printf("***S***\n");
	}
#endif

	// ------------------------------
	// Sort the Estimate Value
	// ------------------------------
	printf("Sorting the estimate Value...\n");
	sortEstIndex = (short **) malloc(sizeof(short *)*noOfItem);


#ifdef DEBUG2
	// Print the sort estimate value
	fprintf(sfilePtr, "\n***Sorted Estimate Index:\n");
	for (i = 0; i < noOfItem; i++)
	{
		for (j = 0; j < noOfItem; j++)
		{
			fprintf(sfilePtr, "%d ", sortEstIndex[i][j]);
		}
		fprintf(sfilePtr, "\n");
	}
#endif


#ifdef M_TIME
	// Mearue Time
	getrusage(RUSAGE_SELF, &start_ru_m3);

	printf("Main Step 2\n");
	printf("------------\n");
	fprintf(timeFilePtr, "Main Step 2(Estimation Set Creation)\n");
	fprintf(timeFilePtr, "-----------------------------------\n");
	printf("***T***Main Step 2(Estimation Set Creation)\n");
	printf("***T***-----------------------------------\n");

	userTimeS = (int) (start_ru_m3.ru_utime.tv_sec - start_ru_m2.ru_utime.tv_sec);
	userTimeM = (int) (start_ru_m3.ru_utime.tv_usec - start_ru_m2.ru_utime.tv_usec);

	sysTimeS = (int) (start_ru_m3.ru_stime.tv_sec - start_ru_m2.ru_stime.tv_sec);
	sysTimeM = (int) (start_ru_m3.ru_stime.tv_usec - start_ru_m2.ru_stime.tv_usec);

	printf("User time: %ds\n", userTimeS);
	printf("User time: %d micros\n", userTimeM);
	fprintf(timeFilePtr, "User time: %ds\n", userTimeS);
	fprintf(timeFilePtr, "User time: %d micros\n", userTimeM);
	printf("***T***User time: %ds\n", userTimeS);
	printf("***T***User time: %d micros\n", userTimeM);

	printf("System time: %ds\n", sysTimeS);
	printf("System time: %d micros\n", sysTimeM);
	fprintf(timeFilePtr, "System time: %ds\n", sysTimeS);
	fprintf(timeFilePtr, "System time: %d micros\n", sysTimeM);
	printf("***T***System time: %ds\n", sysTimeS);
	printf("***T***System time: %d micros\n", sysTimeM);
#endif 

	// ------------------------------
	// Calculate the L_i and H_i
	// ------------------------------
	printf("Calculating the L_i and H_i...\n");


	chooseFlag = (int *) malloc(sizeof(int)*noOfItem);
	for (i = 0; i < noOfItem; i++)
	{
		chooseFlag[i] = 1;
	}


	{
	short *sortEstIndex2;

	int *estValue2;

	int noOfJ;


	// ------------------------------
	// LOOP (ITERATION)
	// ------------------------------
	for (i = 0; i < noOfItem; i++)
	{

			if (i%10 == 0)
			{
				printf(" %d\n", i);
			}

			estValue2 = calculateEst2(i, profit, indCount, freqitemset2, noOfItem, matrix);

#ifdef LARGE_J
			sortEstIndex2 = getWorstJItem2(estValue2, noOfItem, noOfItem-J+1, &noOfJ);
			S[i] = (short *) malloc(sizeof(short)*(noOfItem-J+1));
			{
				int count = 0;
				for (k = 0; k < noOfItem-J+1; k++)
				{
					if (sortEstIndex2[k] != i)
					{
						S[i][count] = sortEstIndex2[k];
						count++;
					}
				}
			}
			transform2Select(s, S[i], noOfJ, sortEstIndex2, i);
			
#else
			sortEstIndex2 = getBestJItem2(estValue2, noOfItem, J, &noOfJ);
			//**** Create the Estimate Set
			S[i] = (short *) malloc(sizeof(short)*J);
			for (k = 0; k < J-1; k++)
			{
				S[i][k] = sortEstIndex2[k];
			}
			S[i][J-1] = (short)i;
#endif

			free(sortEstIndex2);

			free(estValue2);





		if (i%1000 == 0) 
			printf(" %d\n", i);

		L[i] = calculateBenefitFP(data, noOfTrans, matrix, &i, 1, profit, revNo, chooseFlag);


#ifdef LARGE_J

#else

		for (k = 0; k < noOfJ-1; k++)
		{
			s[k] = S[i][k];
		}
		s[noOfJ-1] = i;
#endif



#ifdef DEUBG2
	for (k = 0; k < J; k++)
	{
		fprintf(sfilePtr, "%d ", s[k]);
	}
	fprintf(sfilePtr, "\n");
#endif
		H[i] = calculateBenefitFP2(data, noOfTrans, matrix, s, noOfJ, profit, revNo, chooseFlag, i);
		free(S[i]);
	}
	}

#ifdef DEBUG2
	printf("Writing L_i and H_i into the file....\n");
	lhfilePtr = (FILE *) fopen(lhfilename, "w");
	for (i = 0; i < noOfItem; i++)
	{
		fprintf(lhfilePtr, "%.10f %.10f\n", L[i], H[i]);
	}
#endif


	
	// Checking whether L_i > H_i
	// Make sure L_i <= H_i
	for (i = 0; i < noOfItem; i++)
	{
		if (L[i] > H[i])
		{
			double mid;

			mid = (L[i] + H[i])*0.5;
			L[i] = mid;
			H[i] = mid;
		}
	}
	
	// Create Index for L_i
	sortLIndex = (int *) malloc(sizeof(int)*noOfItem);
	for (i = 0; i < noOfItem; i++)
	{
		sortLIndex[i] = i;
	}

	// Copy L to tempL
	tempL = (double *) malloc(sizeof(double)*noOfItem);
	for (i = 0; i < noOfItem; i++)
	{
		tempL[i] = L[i];
	}

	// Sorting L_i in descending
	for (i = 0; i < noOfItem-1; i++)
	{
		for (j = 0; j < noOfItem-1-i; j++)
		{
			if (tempL[j] < tempL[j+1])
			{
				double temp;
				int temp2;

				// swapping content
				temp = tempL[j];
				tempL[j] = tempL[j+1];
				tempL[j+1] = temp;

				// swapping index
				temp2 = sortLIndex[j];
				sortLIndex[j] = sortLIndex[j+1];
				sortLIndex[j+1] = temp2;
			}
		}
	}

	
#ifdef DEBUG2
	// Print the sorted L_i
	fprintf(sfilePtr, "\n***Sorted L_i is:\n");
	for (i = 0; i < noOfItem; i++)
	{
		fprintf(sfilePtr, "%.5f ", tempL[i]);
	}
	fprintf(sfilePtr, "\n");
#endif

	free(tempL);

	// Setting the J-largest L
	minL = L[sortLIndex[J-1]];
	fprintf(sfilePtr, "\n***minL: %.10f\n", minL);
	printf("***S************ %d **********\n", J);
	printf("***S***\n***S***minL: %.10f\n", minL);

	// Pruning
	// Create Chosen for H_i
	printf("Pruning....\n");
	
	// 0: not selected(i.e. pruned)
	// 1: selected (i.e. not pruned)
	noOfChoose = 0;
	chooseIndex = (int *) malloc(sizeof(int)*noOfItem);
	for (i = 0; i < noOfItem; i++)
	{
		if (H[i] < minL)
		{
			chooseFlag[i] = 0;
		}
		else
		{
			chooseFlag[i] = 1;

			chooseIndex[noOfChoose] = i;
			noOfChoose++;
		}
	}


#ifdef M_TIME
	// Mearue Time
	getrusage(RUSAGE_SELF, &start_ru_m4);

	printf("Main Step 3\n");
	printf("------------\n");
	fprintf(timeFilePtr, "Main Step 3(Item Pruning)\n");
	fprintf(timeFilePtr, "-----------------------------------\n");
	printf("***T***Main Step 3(Item Pruning)\n");
	printf("***T***-----------------------------------\n");

	userTimeS = (int) (start_ru_m4.ru_utime.tv_sec - start_ru_m3.ru_utime.tv_sec);
	userTimeM = (int) (start_ru_m4.ru_utime.tv_usec - start_ru_m3.ru_utime.tv_usec);

	sysTimeS = (int) (start_ru_m4.ru_stime.tv_sec - start_ru_m3.ru_stime.tv_sec);
	sysTimeM = (int) (start_ru_m4.ru_stime.tv_usec - start_ru_m3.ru_stime.tv_usec);

	printf("User time: %ds\n", userTimeS);
	printf("User time: %d micros\n", userTimeM);
	fprintf(timeFilePtr, "User time: %ds\n", userTimeS);
	fprintf(timeFilePtr, "User time: %d micros\n", userTimeM);
	printf("***T***User time: %ds\n", userTimeS);
	printf("***T***User time: %d micros\n", userTimeM);

	printf("System time: %ds\n", sysTimeS);
	printf("System time: %d micros\n", sysTimeM);
	fprintf(timeFilePtr, "System time: %ds\n", sysTimeS);
	fprintf(timeFilePtr, "System time: %d micros\n", sysTimeM);
	printf("***T***System time: %ds\n", sysTimeS);
	printf("***T***System time: %d micros\n", sysTimeM);
#endif 

	// Print the Selected Item
	fprintf(sfilePtr, "\n***The %d selected items not pruned are:\n", noOfChoose);
	printf("***S***\n***S***The %d selected items not pruned are:\n", noOfChoose);
	for (i = 0; i < noOfItem; i++)
	{
		if (chooseFlag[i] == 1)
		{
			fprintf(sfilePtr, "%d ", i);
			printf("%d ", i);
		}
	}
	fprintf(sfilePtr, "\n");
	printf("\n");

	// 
	fprintf(afilePtr, "\n*** J = %d\n", J);
	fprintf(afilePtr, "*** no. of not pruned = %d\n", noOfChoose);
	printf("\n*** J = %d\n", J);
	printf("*** no. of not pruned = %d\n", noOfChoose);
	

	b = (double *) malloc(sizeof(double)*noOfItem);

//#if 0
	isRemove= 0;
	indexB = 0;

#ifdef M_TIME
	// Mearue Time
	getrusage(RUSAGE_SELF, &start_ru_m5);
#endif

	if (noOfChoose > J)
	{
	do
	{
		// Calculate the benefit 
		printf("Calculating the benefit (no. of choose: %d)...\n", noOfChoose);
		
		fprintf(sfilePtr, "\n***Benefit (no. of choose: %d) is:\n", noOfChoose);
		printf("***S***\n***S******Benefit (no. of choose: %d) is:\n", noOfChoose);
		
		for (i = 0; i < noOfChoose; i++)
		{
			int index_i = chooseIndex[i];
			if (chooseFlag[index_i] == 1)
			{
				if (isRemove == 1)
				{
					// Check whether the last removal item is in the set or not
					int isInTheSet = 0;

#if LARGE_J
					isInTheSet = 1;
					for (j = 0; j < noOfEltS[index_i]; j++)
					{
						if (S[index_i][j] == indexB)
						{
							isInTheSet = 0;
							break;
						}
					}
#else
					isInTheSet = 0;
					for (j = 0; j < noOfEltS[index_i]; j++)
					{
						if (S[index_i][j] == indexB)
						{
							isInTheSet = 1;
							break;
						}
					}
#endif

					// if the last removal item is in the set (need calculate the b)
					if (isInTheSet == 1)
					{
						// 
						// Create the Estimate Set Again


#ifdef LARGE_J
						// itemAdd
						itemAdd = S[index_i][noOfEltS[index_i]-1];

						// for all not choose
						noOfEltS[index_i] = 0;
						for (k = 0; k < noOfItem; k++)
						{
							if (chooseFlag[sortEstIndex[index_i][k]] == 0)
							{
								tempNo = sortEstIndex[index_i][k];
								S[index_i][noOfEltS[index_i]] = tempNo;
								noOfEltS[index_i]++;

								// infl
								infl[tempNo]++;
							}
						
						}

						// for some choose
						if (noOfEltS[index_i] < noOfItem-J)
						{
							int tempK;
							for (k = 0; k < noOfItem; k++)
							{
								tempK = noOfItem-1-k;
								// chosen item which is the (k+1)-largest
								if ((chooseFlag[sortEstIndex[index_i][tempK]] == 1) && (sortEstIndex[index_i][tempK] != index_i))
								{
									tempNo = sortEstIndex[index_i][tempK];
									S[index_i][noOfEltS[index_i]] = tempNo;
									noOfEltS[index_i]++;

									// infl
									infl[tempNo]++;

									
						
									if (noOfEltS[index_i] >= noOfItem-J)
									{
										
										break;
									}
								}
							}
						
						}

						


						transform2Select(s, S[index_i], J, S[index_i], index_i);


						// diff
						b[index_i] = b[index_i] + calculateBenefitFP_diff(data, noOfTrans, matrix, s, J, profit, revNo, chooseFlag, itemAdd, itemRemove);
#else
						noOfEltS[index_i] = 0;
						for (k = 0; k < noOfItem; k++)
						{
							// chosen item which is the (k+1)-largest
							if ((chooseFlag[sortEstIndex[index_i][k]] == 1)&& (sortEstIndex[index_i][k] != index_i))
							{
								tempNo = sortEstIndex[index_i][k];
								S[index_i][noOfEltS[index_i]] = tempNo;
								noOfEltS[index_i]++;

								// infl
								infl[tempNo]++;

								if (noOfEltS[index_i] == J-1)
								{
									break;
								}
							}
						}
						//
						S[index_i][noOfEltS[index_i]] = index_i;
						noOfEltS[index_i]++;

						// I think itemAdd is the last one (correct)
						itemAdd = S[index_i][J-2];

						for (k = 0; k < J; k++)
						{
							s[k] = S[index_i][k];
						}
						//b[index_i] = calculateBenefitFP(data, noOfTrans, matrix, s, noOfEltS[index_i], profit, revNo, chooseFlag);			

						// diff
						b[index_i] = b[index_i] + calculateBenefitFP_diff(data, noOfTrans, matrix, s, noOfEltS[index_i], profit, revNo, chooseFlag, itemAdd, itemRemove);
#endif				
					
						
					}
				}
				else
				{
					//
					// Create the Estimate Set Again

					// sorting again
					//for (k = 0; k < noOfChoose; k++)
					//{
						int *estValue2;

						int *tempPtr;

						//int index_k = chooseIndex[k];



						estValue2 = calculateEst2(index_i, profit, indCount, freqitemset2, noOfItem, matrix);

						tempPtr = sort(estValue2, noOfItem);
						sortEstIndex[index_i] = (short *) malloc(sizeof(short)*noOfItem);
						for (j = 0; j < noOfItem; j++)
						{
							sortEstIndex[index_i][j] = (short) tempPtr[j];
						}
						free(estValue2);
						free(tempPtr);
					//}

#ifdef LARGE_J
					S[index_i] = (short *) malloc(sizeof(short)*(noOfItem-J+1));

					// for all not choose
					noOfEltS[index_i] = 0;
					for (k = 0; k < noOfItem; k++)
					{
						if (chooseFlag[sortEstIndex[index_i][k]] == 0)
						{
							tempNo = sortEstIndex[index_i][k];
							S[index_i][noOfEltS[index_i]] = tempNo;
							noOfEltS[index_i]++;

							// infl
							infl[tempNo]++;
						}
						
					}

					// for some choose
					if (noOfEltS[index_i] < noOfItem-J)
					{
						int tempK;
						for (k = 0; k < noOfItem; k++)
						{
							tempK = noOfItem-1-k;
							// chosen item which is the (k+1)-largest
							if ((chooseFlag[sortEstIndex[index_i][tempK]] == 1) && (sortEstIndex[index_i][tempK] != index_i))
							{
								tempNo = sortEstIndex[index_i][tempK];
								S[index_i][noOfEltS[index_i]] = tempNo;
								noOfEltS[index_i]++;

								// infl
								infl[tempNo]++;
						
								if (noOfEltS[index_i] >= noOfItem-J)
								{
									break;
								}
							}
						}
						
					}

					transform2Select(s, S[index_i], J, S[index_i], index_i);

					b[index_i] = calculateBenefitFP(data, noOfTrans, matrix, s, J, profit, revNo, chooseFlag);			
#else
					S[index_i] = (short *) malloc(sizeof(short)*J);

					//
					noOfEltS[index_i] = 0;
					for (k = 0; k < noOfItem; k++)
					{
						// chosen item which is the (k+1)-largest
						if ((chooseFlag[sortEstIndex[index_i][k]] == 1) && (sortEstIndex[index_i][k] != index_i))
						{
							tempNo = sortEstIndex[index_i][k];
							S[index_i][noOfEltS[index_i]] = tempNo;
							noOfEltS[index_i]++;

							// infl
							infl[tempNo]++;
						
							if (noOfEltS[index_i] == J-1)
							{
								break;
							}
						}
					}
					//s[noOfEltS[index_i]] = index_i;
					S[index_i][noOfEltS[index_i]] = index_i;
					noOfEltS[index_i]++;

					for (k = 0; k < J; k++)
					{
						s[k] = S[index_i][k];
					}
					b[index_i] = calculateBenefitFP(data, noOfTrans, matrix, s, noOfEltS[index_i], profit, revNo, chooseFlag);			
#endif				

					

				}

#ifdef DEBUG2
fprintf(sfilePtr, " %d : b[i] : %.2f\n", index_i, b[index_i]);
#endif
#ifdef DEBUG2
fprintf(sfilePtr, "%d - ", index_i);
for (k = 0; k < J; k++)
{
	fprintf(sfilePtr, "%d ", s[k]);
}
fprintf(sfilePtr, "\n");
#endif


			}
		}

		
		// Print the set S i
		
#ifdef DEBUG2
		fprintf(sfilePtr, "\n***The set S_i is:\n");
		for (i = 0; i < noOfItem; i++)
		{
			if (chooseFlag[i] == 1)
			{
				fprintf(sfilePtr, "%d %d ", i, noOfEltS[i]);
				for (j = 0; j < noOfEltS[i]; j++)
				{
					fprintf(sfilePtr, "%d ", S[i][j]);
				}
				fprintf(sfilePtr, "\n");
			}
		}
#endif

//goto hello;
		

		// Print no. of Non-pruned items
		
		noOfNonZero = 0;
		noOfSelect= 0;
		//selectS = (int *) malloc(sizeof(int)*100);
		for (i = 0; i < noOfChoose; i++)
		{
			if (b[chooseIndex[i]] != 0)
			{
				noOfNonZero++;
			}
		}
		printf("*** no. of non-zero = %d\n", noOfNonZero);
		fprintf(afilePtr, "*** no. of non-zero = %d\n", noOfNonZero);


		
	
		
		
		// Find the min. value of b_i
		if (noOfChoose <= 0)
		{
			printf("The no. of remaining <= 0!\n");
			return 0;
		}

		

#ifdef LARGE_J
		for (i = 0; i < noOfChoose; i++)
		{
			infl[chooseIndex[i]] = noOfItem;
		}
		for (i = 0; i < noOfChoose; i++)
		{
			for (j = 0; j < noOfItem-J;j++)
			{
				infl[S[chooseIndex[i]][j]]--;
			}
		}
#else
		for (i = 0; i < noOfChoose; i++)
		{
			infl[chooseIndex[i]] = 0;
		}
		for (i = 0; i < noOfChoose; i++)
		{
			for (j = 0; j < J; j++)
			{
				infl[S[chooseIndex[i]][j]]++;
			}
		}
#endif


		minB = b[chooseIndex[0]];
		indexB = chooseIndex[0];
		indexChooseB = 0;


		{
		double lower;
		double upper;
		int noOfAcc;

		int minInfl;

		int *indexB_list;
		int *indexChooseB_list;

		indexB_list = (int *) malloc(sizeof(int)*noOfItem);
		indexChooseB_list = (int *) malloc(sizeof(int)*noOfItem);

		noOfAcc = 0;
		indexB_list[noOfAcc] = chooseIndex[0];
		indexChooseB_list[noOfAcc] = 0;		
		noOfAcc++;

		lower = minB*0.99;
		upper = minB*1.01;
		for (i = 1; i < noOfChoose; i++)
		{
			if ((lower < b[chooseIndex[i]]) && (b[chooseIndex[i]] < upper))
			{
				// nearly the same
				indexB_list[noOfAcc] = chooseIndex[i];
				indexChooseB_list[noOfAcc] = i;
				noOfAcc++;
			}
			else if (b[chooseIndex[i]] < minB)
			{
				minB = b[chooseIndex[i]];
				//indexB = chooseIndex[i];
				//indexChooseB = i;

				lower = minB*0.99;
				upper = minB*1.01;

				noOfAcc = 0;

				indexB_list[noOfAcc] = chooseIndex[i];
				indexChooseB_list[noOfAcc] = i;
				noOfAcc++;
			}
		}

		// choose the less influence
		minInfl = infl[indexB_list[0]];
		indexB = indexB_list[0];
		indexChooseB = indexChooseB_list[0];
		for (i = 1; i < noOfAcc; i++)
		{
			if (infl[indexB_list[i]] < minInfl)
			{
				minInfl = infl[indexB_list[i]];
				indexB = indexB_list[i];
				indexChooseB = indexChooseB_list[i];
			}
		}

		free(indexB_list);
		free(indexChooseB_list);
		}

#ifdef DEBUG2
fprintf(sfilePtr, "**************min B :%.10f\n", minB);
fprintf(sfilePtr, "**************indexB:%d\n", indexB);
fprintf(sfilePtr, "**************b[indexB]:%.10f\n", b[indexB]);
fprintf(sfilePtr, "**************chooseIndex[noOfChoose-1]:%d\n", chooseIndex[noOfChoose-1]);
for (i = 0; i < noOfChoose; i++)
{
	fprintf(sfilePtr, "%d \n", chooseIndex[i]);
}
fprintf(sfilePtr, "***********\n");
#endif

		// Remove item indexB
		chooseFlag[indexB] = 0;
		free(S[indexB]);
		itemRemove = indexB;
		isRemove = 1;
		for (i = indexChooseB; i < noOfChoose-1; i++)
		{
			chooseIndex[i] = chooseIndex[i+1];
		}
		noOfChoose--;
		
	}
	while(noOfChoose > J);
	}
	
//#endif


#ifdef M_TIME
	// Mearue Time
	getrusage(RUSAGE_SELF, &start_ru_m6);

	printf("Main Step 6\n");
	printf("------------\n");
	fprintf(timeFilePtr, "Main Step 6(Iteration)\n");
	fprintf(timeFilePtr, "-----------------------------------\n");
	printf("***T***Main Step 6(Iteration)\n");
	printf("***T***-----------------------------------\n");

	userTimeS = (int) (start_ru_m6.ru_utime.tv_sec - start_ru_m5.ru_utime.tv_sec);
	userTimeM = (int) (start_ru_m6.ru_utime.tv_usec - start_ru_m5.ru_utime.tv_usec);

	sysTimeS = (int) (start_ru_m6.ru_stime.tv_sec - start_ru_m5.ru_stime.tv_sec);
	sysTimeM = (int) (start_ru_m6.ru_stime.tv_usec - start_ru_m5.ru_stime.tv_usec);

	printf("User time: %ds\n", userTimeS);
	printf("User time: %d micros\n", userTimeM);
	fprintf(timeFilePtr, "User time: %ds\n", userTimeS);
	fprintf(timeFilePtr, "User time: %d micros\n", userTimeM);
	printf("***T***User time: %ds\n", userTimeS);
	printf("***T***User time: %d micros\n", userTimeM);

	printf("System time: %ds\n", sysTimeS);
	printf("System time: %d micros\n", sysTimeM);
	fprintf(timeFilePtr, "System time: %ds\n", sysTimeS);
	fprintf(timeFilePtr, "System time: %d micros\n", sysTimeM);
	printf("***T***System time: %ds\n", sysTimeS);
	printf("***T***System time: %d micros\n", sysTimeM);


	// Mearue Time
	//getrusage(RUSAGE_SELF, &start_ru_m6);

	printf("Overall \n");
	printf("------------\n");
	fprintf(timeFilePtr, "Overall\n");
	fprintf(timeFilePtr, "-----------------------------------\n");
	printf("***T***Overall\n");
	printf("***T***-----------------------------------\n");

	userTimeS = (int) (start_ru_m6.ru_utime.tv_sec - start_ru.ru_utime.tv_sec);
	userTimeM = (int) (start_ru_m6.ru_utime.tv_usec - start_ru.ru_utime.tv_usec);

	sysTimeS = (int) (start_ru_m6.ru_stime.tv_sec - start_ru.ru_stime.tv_sec);
	sysTimeM = (int) (start_ru_m6.ru_stime.tv_usec - start_ru.ru_stime.tv_usec);

	printf("User time: %ds\n", userTimeS);
	printf("User time: %d micros\n", userTimeM);
	fprintf(timeFilePtr, "User time: %ds\n", userTimeS);
	fprintf(timeFilePtr, "User time: %d micros\n", userTimeM);
	printf("***T***User time: %ds\n", userTimeS);
	printf("***T***User time: %d micros\n", userTimeM);

	printf("System time: %ds\n", sysTimeS);
	printf("System time: %d micros\n", sysTimeM);
	fprintf(timeFilePtr, "System time: %ds\n", sysTimeS);
	fprintf(timeFilePtr, "System time: %d micros\n", sysTimeM);
	printf("***T***System time: %ds\n", sysTimeS);
	printf("***T***System time: %d micros\n", sysTimeM);

#endif 	




	// Print the Final Output
	fprintf(sfilePtr, "\n***The Selection Set is :\n");
	printf("***S***\n***S***The Selection Set is :\n");
	totalProfit = 0;
	for (i = 0; i < noOfChoose; i++)
	{ 
		int index_i = chooseIndex[i];
		fprintf(sfilePtr, "%d ", index_i);
		printf("%d ", index_i);
		b[index_i] = calculateBenefit(data, noOfTrans, matrix, chooseIndex, J, index_i, profit[index_i]);
		totalProfit += b[index_i];
	}
	fprintf(sfilePtr, "\n");
	printf("\n");



hello:

	// Print the total Profit
	fprintf(sfilePtr, "\n***The total selection Profit (AMPIS) is : %.10f\n", totalProfit);
	printf("***S***\n***S***The total selection Profit (AMPIS) is : %.10f\n", totalProfit);

	// Calculate the Original Profit
	origTotalProfit = 0;
	for (i = 0; i < noOfItem; i++)
	{
		//origTotalProfit += profit[i]*matrix[i][i];
		origTotalProfit += profit[i]*getMatrix(i, i);
	}
	fprintf(sfilePtr, "\n***The original profit is :%.10f\n", origTotalProfit);
	printf("***S***\n***S***The original profit is :%.10f\n", origTotalProfit);

	// Print the overall
	fprintf(sfilePtr, "\n***The %.2f percent selection(AMPIS) gives %.2f percent profit of the original profit.\n", J/(double)noOfItem*100, totalProfit/origTotalProfit*100);
	printf("***S***\n***S******The %.2f percent selection(AMPIS) gives %.2f percent profit of the original profit.\n", J/(double)noOfItem*100, totalProfit/origTotalProfit*100);


	

	printf("Freeing memeory...\n");

 /* free memory ------------------------------------*/
 printf("\ndestroy\n");
 destroy();

/*
	// free the data
	for (i = 0; i < noOfTrans; i++)
	{
		free(data[i]->item);
		free(data[i]);
	}
	free(data);



	// free the individual count
	free(indCount);
	


	// free the sor estimate index
	for (i = 0; i < noOfItem; i++)
	{
		free(sortEstIndex[i]);
	}
	free(sortEstIndex);

	//
	free(L);
	free(H);
	free(s);

	// 
	free(chooseIndex);
	free(chooseFlag);

	//
	free(b);

	fclose(lhfilePtr);
	fclose(sfilePtr);
	fclose(mfilePtr);
	fclose(afilePtr);
	fclose(profFilePtr);
*/
	return 0;
}

/*
 * Function  : readData
 * Purpose   : to read the data file into memory
 * Parameters: 
 *   1. filename: the filename storing the data
 *   2. noOfTrans: no. of trans 
 */
transPtr *readData(STRING filename, int noOfTrans)
{
	FILE *filePtr;
	transPtr *returnValue;
	int transNo;

	int i;

	filePtr = (FILE *) fopen(filename, "r");
	if (filePtr == NULL)
	{
		printf("File %s cannot be opened!\n", filename);
		exit(0);
	}

	returnValue = (transPtr *) malloc(sizeof(transPtr)*noOfTrans);

	transNo = 0;

	while (!feof(filePtr))
	{
		returnValue[transNo] = (transPtr) malloc(sizeof(trans));

		fscanf(filePtr, "%d", &(returnValue[transNo]->noOfItem));

		returnValue[transNo]->item = (int *) malloc(sizeof(int)*returnValue[transNo]->noOfItem);

		for (i = 0; i < returnValue[transNo]->noOfItem; i++)
		{
			fscanf(filePtr, " %d", &(returnValue[transNo]->item[i]));
		}

		fscanf(filePtr, "\n");

		transNo++;
	}

	fclose(filePtr);

	return returnValue;
}


/*
 * Function  : incCount
 * Purpose   : to increment the count of the items in the transaction
 * Parameters: 
 *   1. indCount - the count stored for the items
 *   2. transaction - the transaction
 *   3. freqItemset - the freq. itemset stored
 *   4. noOfItem - no. of items in the database
 */
// increment the individual count of some items for a transaction 
void incCount(int *indCount, transPtr transaction, largeitemPtr *freqitemset, int noOfItem)
{
	int i, j, k;
	int *transFlag;

	int freqitemSize = 1; // the estimated freq. set size 2


	transFlag = (int *) malloc(sizeof(int)*transaction->noOfItem);

	// Set the Flags as 1
	for (i = 0; i < transaction->noOfItem; i++)
	{
		transFlag[i] = 1;
	}

	// for each item in the transaction
	for (i = 0; i < transaction->noOfItem; i++)
	{
		// for each item in the frequent itemset
		for (j = 0; j < freqitemset[freqitemSize]->noOfItemSet; j++)
		{
			if (transaction->item[i] == freqitemset[freqitemSize]->item[j]->itemNo[0])
			{
				for (k = i+1; k < transaction->noOfItem; k++)
				{
					if (transaction->item[k] == freqitemset[freqitemSize]->item[j]->itemNo[1])
					{
						transFlag[i] = 0;
						transFlag[k] = 0;
					}
				}
			}

			//*** can break the loop if the frequent item set is sorted among the size 2 set
		}
	}

	for (i = 0; i < transaction->noOfItem; i++)
	{
		if (transFlag[i] == 1)
		{
			indCount[transaction->item[i]]++;
		}
	}

	free(transFlag);
}

/*
 * Function  : sortFreqItemSet
 * Purpose   : to sort the freq. itemset of size 2 ordering according the item no. (in ascending order)
 * Parameters: 
 *   1. freqItemset : frequent itemset
 */
void sortFreqItemSet(largeitemPtr *freqitemset)
{
	int i;
	
	int freqItemSize = 1;

	for (i = 0; i < freqitemset[freqItemSize]->noOfItemSet; i++)
	{
		if (freqitemset[freqItemSize]->item[i]->itemNo[0] > freqitemset[freqItemSize]->item[i]->itemNo[1])
		{
			int temp;

			// swapping
			temp = freqitemset[freqItemSize]->item[i]->itemNo[0];
			freqitemset[freqItemSize]->item[i]->itemNo[0] = freqitemset[freqItemSize]->item[i]->itemNo[1];
			freqitemset[freqItemSize]->item[i]->itemNo[1] = temp;
		}
	}
}

/*
 * Function  : deriveIndividualCount
 * Purpose   : to derive the individual count
 * Parameters: 
 *   1. data: the data in the database
 *   2. freqitemset: freq. itemset
 *   3. noOfTrans: no. of transactions
 *   4. noOfItem: no. of items
 */
int *deriveIndividualCount(transPtr *data, largeitemPtr *freqitemset, int noOfTrans, int noOfItem)
{
	int *returnValue;

	int i;



	returnValue = (int *) malloc(sizeof(int)*noOfItem);

	// Initialize as zero fo the returnValue
	for (i = 0; i < noOfItem; i++)
	{
		returnValue[i] = 0;
	}

	// sort the f-tree for size 2
	sortFreqItemSet(freqitemset);


	for (i = 0; i < noOfTrans; i++)
	{
		//incCount(returnValue, data[i], freqitemset, noOfItem);

		if (data[i]->noOfItem == 1)
		{
			returnValue[data[i]->item[0]]++;
		}

	}

	return returnValue;
}



/*
 * Function  : getSupportTwo
 * Purpose   : to get the support count of 2-itemsets containing item i and item j (where i < j)
 * Parameters: 
 *   1. freqitemset: freq. itemset
 *   2. i, j: item nos. of 2-itemsets
 */
int getSupportTwo(largeitemPtr *freqitemset, int i, int j)
{
	int returnValue = 0;
	int freqItemSize = 1;

	sortFreqItemSet(freqitemset);

	if (i > j)
	{
		int temp;

		temp = i;
		i = j;
		j = temp;
	}

	for (i = 0; i < freqitemset[freqItemSize]->noOfItemSet; i++)
	{
		if ((freqitemset[freqItemSize]->item[i]->itemNo[0] == i) && (freqitemset[freqItemSize]->item[i]->itemNo[1] == j))
		{
			returnValue = freqitemset[freqItemSize]->item[i]->freq;
			break;
		}
	}

	return returnValue;
}

/*
 * Function  : isContain
 * Purpose   : to find wheter the transaction contains a set of elements
 * Parameters: 
 *   1. transaction: a transaction
 *   2. elt: a set of elements
 *   3. noOfElt: no. of elements
 * NOTE:
 *   1. return:
 *      1 :true
 *      0 :false
 *   2. Asssume the set of elements is sorted
 */

int isContain(transPtr transaction, int *elt, int noOfElt)
{
	int i;

	int curr = 0;

	for (i = 0; i < transaction->noOfItem; i++)
	{
		
		if (elt[curr] == transaction->item[i])
		{
			curr++;

			if (curr == noOfElt)
			{
				return 1;
			}
		}
		else if (elt[curr] < transaction->item[i])
		{
			return 0;
		}
	}

	return 0;
}

/*
 * Function  : isContainAny
 * Purpose   : 
 * Parameters: 
 *   1. transaction: a transaction
 *   2. isChoose: a list of items indicating whether it is chosen or not
 * NOTE:
 *   1. return:
 *        1:true
 *        0:false
 */
int isContainAny(transPtr transaction, int *isChoose)
{
	int returnValue = 0;

	int i;

	for (i = 0; i < transaction->noOfItem; i++)
	{
		if (isChoose[transaction->item[i]] == 1)
		{
			returnValue = 1;
			return returnValue;
		}
	}
	return returnValue;
}

/*
 * Function  : isContainAny
 * Purpose   : 
 * Parameters: 
 *   1. transaction: a transaction
 *   2. isChoose: a list of items indicating whether it is chosen or not
 *   3. itemSpecial: item no. which is in the left hand side of the rule
 * NOTE:
 *   1. return:
 *        2 :true (contain itemSpecial)
 *        1 :true (not contain itemSpecial)
 *        0 :false
 */
int isContainAny2(transPtr transaction, int *isChoose, int itemSpecial)
{
	int returnValue = 0;

	int i;

	int itemNo;

	for (i = 0; i < transaction->noOfItem; i++)
	{
		itemNo = transaction->item[i];
		if (isChoose[itemNo] == 1)
		{
			returnValue = 1;
			
			if (itemNo == itemSpecial)
			{
				return 2;
			}
		}
	}
	return returnValue;
}


/*
 * Function  : getSupport
 * Purpose   : to get the support of itemsets of any size from the database
 * Parameters: 
 *   1. data: the data in the database
 *   2. noOfTrans: no. of transactions
 *   3. elt: the set of elements
 *   4. noOfElt: no. of elements in the set
 * NOTE:
 *   1. the elements in the set is sorted in ascending order
 *   2. the elements in the set is updated/changed.
 */
int getSupport(transPtr *data, int noOfTrans, int *elt, int noOfElt)
{
	int returnValue = 0;
	int i, j;

	// sort the elt
	for (i = 0; i < noOfElt-1; i++)
	{
		for (j = 0; j < noOfElt-1-i; j++)
		{
			if (elt[j] > elt[j+1])
			{
				int temp;

				//swapping
				temp = elt[j];
				elt[j] = elt[j+1];
				elt[j+1] = elt[j];
			}
		}
	}

	for (i = 0; i < noOfTrans; i++)
	{
		if (isContain(data[i], elt, noOfElt) == 1)
		{
			returnValue++;
		}
	}
	
	

	return returnValue;
}

/*
 * Function  : calculateEst
 * Purpose   : to calculate the estimated values in the algorithm
 * Parameters: 
 *   1. profit: profit of each item
 *   2. indCount: the individual count of each item
 *   3. freqitemset: freq. itemsets
 *   4. noOfItem: no. of items
 *   5. matrix: a matrix to store the count of co-occurence
 */
int **calculateEst(int *profit, int *indCount, largeitemPtr *freqitemset, int noOfItem, short **matrix)
{
	int **returnValue;

	int i, j;

	returnValue = (int **) malloc(sizeof(int)*noOfItem);
	for (i = 0; i < noOfItem; i++)
	{
		returnValue[i] = (int *) malloc(sizeof(int)*noOfItem);
	}

	for (i = 0; i < noOfItem; i++)
	{
		for (j = 0; j < noOfItem; j++)
		{
			if (i != j)
			{
				int no1, no2;
				no1 = i;
				no2 = j;

				// make no1 < no2
				if (no1 > no2)
				{
					int temp;

					//swapping
					temp = no1;
					no1 = no2;
					no2 = temp;
				}

				//returnValue[i][j] = profit[j]*indCount[j] + (profit[j] + profit[i])*getSupportTwo(freqitemset, i, j);
				//returnValue[i][j] = profit[j]*indCount[j] + (profit[j] + profit[i])*matrix[no1][no2];
				returnValue[i][j] = profit[j]*indCount[j] + (profit[j] + profit[i])*getMatrix(no1,no2);
			}
			else
			{
				returnValue[i][i] = profit[i]*indCount[i];
			}
		}
	}

	return returnValue;
}

/*
 * Function  : calculateEst2
 * Purpose   : to calculate the estimated values in the algorithm
 * Parameters: 
 *   1. profit: profit of each item
 *   2. indCount: the individual count of each item
 *   3. freqitemset: freq. itemsets
 *   4. noOfItem: no. of items
 *   5. matrix: a matrix to store the count of co-occurence
 */
int *calculateEst2(int itemI, int *profit, int *indCount, largeitemPtr *freqitemset, int noOfItem, short **matrix)
{
	int *returnValue;

	int temp;

	int i, j;
	int no1, no2;

	int noOfExist;

	returnValue = (int *) malloc(sizeof(int)*noOfItem);

	i = itemI;

	//for (i = 0; i < noOfItem; i++)
	//{
		for (j = 0; j < noOfItem; j++)
		{
			if (i != j)
			{
				
				no1 = i;
				no2 = j;

				// make no1 < no2
				if (no1 > no2)
				{
					//int temp;

					//swapping
					temp = no1;
					no1 = no2;
					no2 = temp;
				}

				noOfExist = getMatrix(no1,no2);

				

				//returnValue[i][j] = profit[j]*indCount[j] + (profit[j] + profit[i])*getSupportTwo(freqitemset, i, j);
				//returnValue[j] = profit[j]*indCount[j] + (profit[j] + profit[i])*matrix[no1][no2];
				//returnValue[j] = profit[j]*indCount[j] + (profit[j] + profit[i])*noOfExist;

				//returnValue[j] = profit[j]*getMatrix(no1, no2);

				temp = 0;
				if (indCount[j] != 0)
				{
					temp += profit[j]*indCount[j];
				}
				if (noOfExist != 0)
				{
					temp += (profit[j] + profit[i])*noOfExist;
				}
				returnValue[j] = temp;
			}
			else
			{
				//returnValue[i] = profit[i]*indCount[i];

				//returnValue[i] = -1;

				temp = 0;
				if (indCount[i] != 0)
				{
					temp += profit[i]*indCount[i];
				}
				returnValue[i] = temp;
				
			}
		}
	//}

	return returnValue;
}




/*
 * Function  : sort
 * Purpose   : to sort the element in the set
 * Parameters: 
 *   1. element: the elements in the set
 *   2. noOfElt: no. of elements in the set
 * NOTE:
 *   1. The output will be sorted in descending order
 *   2. The output is the index sorted
 *   3. The sort is quick sort.
 */
int *sort(int *element, int noOfElt)
{
	int *returnValue;

	int i;//, j;

	//int *tempElt;

	int noOfNonZero;

	int startIndex, endIndex;

	returnValue = (int *) malloc(sizeof(int)*noOfElt);
	//tempElt = (int *) malloc(sizeof(int)*noOfElt);
	for (i = 0; i < noOfElt; i++)
	{
		returnValue[i] = i;
		//tempElt[i] = element[i];
	}

	// Check for zero elements
	startIndex = -1;
	endIndex = noOfElt;
	while (1)
	{
		do
		{
			startIndex++;
		}
		while ((startIndex < noOfElt) && (element[startIndex] != 0));

		do
		{
			endIndex--;
		}
		while ((endIndex >= 0) && (element[endIndex] == 0));

		// swap
		if (startIndex < endIndex)
		{
			//void q_swap_i(int *sorted_index, int *value, int i, int j)
			q_swap_i(returnValue, element, startIndex, endIndex);
		}
		else
		{
			break;
		}
	}
	
	// here, all zero's are put at the end of the list

	// find the end elt (which is not zero)
	noOfNonZero = noOfElt;
	for (i = 0; i < noOfElt; i++)
	{
		if (element[i] == 0)
		{
			noOfNonZero = i+1;
			break;
		}
	}
	
	//void quicksort(int *sorted_index, double *value, int start_index, int end_index)
	//quicksort_i(returnValue, element, 0, noOfElt-1);
	quicksort_i(returnValue, element, 0, noOfNonZero-1);

/*
	for (i = 0; i < noOfElt-1; i++)
	{
		for (j = 0; j < noOfElt-1-i; j++)
		{
			if (element[j] < element[j+1])
			{
				int temp;

				//swap content
				temp = element[j+1];
				element[j+1] = element[j];
				element[j] = temp;

				//swap index
				temp = returnValue[j+1];
				returnValue[j+1] = returnValue[j];
				returnValue[j] = temp;
			}
		}
	}
*/
	//free(tempElt);

	return returnValue;
}

/*
 * Function  : getBestJItem
 * Purpose   : to sort the J largest values of items in descending order 
 * Parameters: 
 *   1. element: the elements in the set
 *   2. noOfElt: no. of elements in the set
 *   3. boundJ: the best J number
 * NOTE:
 *   1. The output will be sorted in descending order
 *   2. The output is the index sorted
 */
short *getBestJItem(int *element, int noOfElt, int boundJ)
{
	short *returnValue;
	short *returnValue2;

	int i, j;

	int index, value;
	short value2;

	int noOfNonZero;
	int startIndex, endIndex;


	returnValue = (short *) malloc(sizeof(short)*noOfElt);
	for (i = 0; i < noOfElt; i++)
	{
		returnValue[i] = (short)i;
	}

	returnValue2 = (short *) malloc(sizeof(short)*boundJ);
	for (i = 0; i < boundJ; i++)
	{
		returnValue2[i] = (short)i;
	}

	//returnValue = (short *) malloc(sizeof(short)*noOfElt);
	//for (i = 0; i < noOfElt; i++)
	//{
	//	returnValue[i] = (short)i;
	//}

	// print the element
	//printf("The element is:\n");
	//for (i = 0; i < noOfElt; i++)
	//{
	//	printf("%d ", element[i]);
	//}//
	//printf("\n");





	// Check for zero elements
	startIndex = -1;
	endIndex = noOfElt;
	while (1)
	{
		do
		{
			startIndex++;
		}
		while ((startIndex < noOfElt) && (element[startIndex] != 0));

		do
		{
			endIndex--;
		}
		while ((endIndex >= 0) && (element[endIndex] == 0));

		// swap
		if (startIndex < endIndex)
		{
			//void q_swap_i(int *sorted_index, int *value, int i, int j)
			q_swap_i_indShort(returnValue, element, startIndex, endIndex);
		}
		else
		{
			break;
		}
	}
	
	// here, all zero's are put at the end of the list

	// find the end elt (which is not zero)
	noOfNonZero = noOfElt;
	for (i = 0; i < noOfElt; i++)
	{
		if (element[i] == 0)
		{
			noOfNonZero = i+1;
			break;
		}
	}

	

	// sort for the first boundJ elements
/*	for (i = 0; i < boundJ-1; i++)
	{
		for (j = 0; j < boundJ-1-i; j++)
		{
			if (element[j] < element[j+1])
			{
				int temp;
				short temp2;

				// swap content
				temp = element[j+1];
				element[j+1] = element[j];
				element[j] = temp;

				// swap index
				temp2 = returnValue[j+1];
				returnValue[j+1] = returnValue[j];
				returnValue[j] = temp2;
			}
		}
	}
*/
	if (noOfNonZero < boundJ)
	{
		quicksort_i_indShort(returnValue, element, 0, noOfNonZero-1);
	}
	else
	{
		//void quicksort_i_indShort(short *sorted_index, int *value, int start_index, int end_index)
		quicksort_i_indShort(returnValue, element, 0, boundJ-1);
	

		// check the remaining elts are the boundJ largest
		for (i = boundJ; i < noOfNonZero; i++)
		{
			if (element[i] > element[boundJ-1])
			{
				for (j = 0; j < boundJ; j++)
				{
					if (element[i] > element[j])
					{
						index = j;
						value = element[i];
						value2 = returnValue[i];
						//value2 = (short)i;
						break;
					}
				}

				// move content
				element[i] = element[boundJ-1];
				for (j = boundJ-1; j > index; j--)
				{
					element[j] = element[j-1];
				}
				element[index] = value;

				// move index
				//returnValue[i] = returnValue[boundJ-1];
				for (j = boundJ-1; j > index; j--)
				{
					returnValue[j] = returnValue[j-1];
				}
				returnValue[index] = value2;
			}
		}

		// print sorted index
		//printf("Sorted index:\n");
		//for (i = 0; i < boundJ; i++)
		//{
		//	printf("%d ", returnValue[i]);
		//}
		//printf("\n");
	}

	memcpy(returnValue2, returnValue, sizeof(short)*boundJ);

	free(returnValue);
	
	return returnValue2;
}


/*
 * Function  : getBestJItem2
 * Purpose   : to sort the J largest values of items in descending order 
 * Parameters: 
 *   1. element: the elements in the set
 *   2. noOfElt: no. of elements in the set
 *   3. boundJ: the best J number
 * NOTE:
 *   1. The output will be sorted in descending order
 *   2. The output is the index sorted
 */

short *getBestJItem2(int *element, int noOfElt, int boundJ, int *noOfJ)
{
	short *returnValue;
	short *returnValue2;

	int i, j;

	int index, value;
	short value2;

	int noOfNonZero;
	int startIndex, endIndex;


	returnValue = (short *) malloc(sizeof(short)*noOfElt);
	for (i = 0; i < noOfElt; i++)
	{
		returnValue[i] = (short)i;
	}

	returnValue2 = (short *) malloc(sizeof(short)*boundJ);
	for (i = 0; i < boundJ; i++)
	{
		returnValue2[i] = (short)i;
	}

	//returnValue = (short *) malloc(sizeof(short)*noOfElt);
	//for (i = 0; i < noOfElt; i++)
	//{
	//	returnValue[i] = (short)i;
	//}

	// print the element
	//printf("The element is:\n");
	//for (i = 0; i < noOfElt; i++)
	//{
	//	printf("%d ", element[i]);
	//}//
	//printf("\n");





	// Check for zero elements
	startIndex = -1;
	endIndex = noOfElt;
	while (1)
	{
		do
		{
			startIndex++;
		}
		while ((startIndex < noOfElt) && (element[startIndex] != 0));

		do
		{
			endIndex--;
		}
		while ((endIndex >= 0) && (element[endIndex] == 0));

		// swap
		if (startIndex < endIndex)
		{
			//void q_swap_i(int *sorted_index, int *value, int i, int j)
			q_swap_i_indShort(returnValue, element, startIndex, endIndex);
		}
		else
		{
			break;
		}
	}
	
	// here, all zero's are put at the end of the list

	// find the end elt (which is not zero)
	noOfNonZero = noOfElt;
	for (i = 0; i < noOfElt; i++)
	{
		if (element[i] == 0)
		{
			noOfNonZero = i+1;
			break;
		}
	}

	

	// sort for the first boundJ elements
/*	for (i = 0; i < boundJ-1; i++)
	{
		for (j = 0; j < boundJ-1-i; j++)
		{
			if (element[j] < element[j+1])
			{
				int temp;
				short temp2;

				// swap content
				temp = element[j+1];
				element[j+1] = element[j];
				element[j] = temp;

				// swap index
				temp2 = returnValue[j+1];
				returnValue[j+1] = returnValue[j];
				returnValue[j] = temp2;
			}
		}
	}
*/
	if (noOfNonZero < boundJ)
	{
		quicksort_i_indShort(returnValue, element, 0, noOfNonZero-1);
	}
	else
	{
		//void quicksort_i_indShort(short *sorted_index, int *value, int start_index, int end_index)
		quicksort_i_indShort(returnValue, element, 0, boundJ-1);
	

		// check the remaining elts are the boundJ largest
		for (i = boundJ; i < noOfNonZero; i++)
		{
			if (element[i] > element[boundJ-1])
			{
				for (j = 0; j < boundJ; j++)
				{
					if (element[i] > element[j])
					{
						index = j;
						value = element[i];
						value2 = returnValue[i];
						//value2 = (short)i;
						break;
					}
				}

				// move content
				element[i] = element[boundJ-1];
				for (j = boundJ-1; j > index; j--)
				{
					element[j] = element[j-1];
				}
				element[index] = value;

				// move index
				//returnValue[i] = returnValue[boundJ-1];
				for (j = boundJ-1; j > index; j--)
				{
					returnValue[j] = returnValue[j-1];
				}
				returnValue[index] = value2;
			}
		}

		// print sorted index
		//printf("Sorted index:\n");
		//for (i = 0; i < boundJ; i++)
		//{
		//	printf("%d ", returnValue[i]);
		//}
		//printf("\n");
	}

	memcpy(returnValue2, returnValue, sizeof(short)*boundJ);

	free(returnValue);
	
	
	(*noOfJ) = boundJ;
	for (i = 0; i < boundJ; i++)
	{
		if (element[i] == 0)
		{
			(*noOfJ) = i+1;
			break;
		}
	}
	
	return returnValue2;
}



/*
 * Function  : getWorstJItem2 (dummy)
 * Purpose   : to sort the J largest values of items in descending order 
 * Parameters: 
 *   1. element: the elements in the set
 *   2. noOfElt: no. of elements in the set
 *   3. boundJ: the best J number
 *   4. noOfJ: return value (no. of return values)
 * NOTE:
 *   1. The output will be sorted in descending order
 *   2. The output is the index sorted
 */

short *getWorstJItem2(int *element, int noOfElt, int boundJ, int *noOfJ)
{
	short *returnValue;
	short *returnValue2;

	int i, j;

	int index, value;
	short value2;

	int noOfZero, noOfNonZero;
	int startIndex, endIndex;


	returnValue = (short *) malloc(sizeof(short)*noOfElt);
	for (i = 0; i < noOfElt; i++)
	{
		returnValue[i] = (short)i;
	}

	returnValue2 = (short *) malloc(sizeof(short)*boundJ);
	for (i = 0; i < boundJ; i++)
	{
		returnValue2[i] = (short)i;
	}

	//returnValue = (short *) malloc(sizeof(short)*noOfElt);
	//for (i = 0; i < noOfElt; i++)
	//{
	//	returnValue[i] = (short)i;
	//}

	// print the element
	//printf("The element is:\n");
	//for (i = 0; i < noOfElt; i++)
	//{
	//	printf("%d ", element[i]);
	//}//
	//printf("\n");





	// Check for zero elements
	startIndex = -1;
	endIndex = noOfElt;
	while (1)
	{
		do
		{
			startIndex++;
		}
		while ((startIndex < noOfElt) && (element[startIndex] == 0));

		do
		{
			endIndex--;
		}
		while ((endIndex >= 0) && (element[endIndex] != 0));

		// swap
		if (startIndex < endIndex)
		{
			//void q_swap_i(int *sorted_index, int *value, int i, int j)
			q_swap_i_indShort(returnValue, element, startIndex, endIndex);
		}
		else
		{
			break;
		}
	}
	
	// here, all zero's are put at the beginning of the list

	// find the end elt (which is not zero)
	noOfZero = noOfElt;
	for (i = noOfElt-1; i >= 0; i--)
	{
		if (element[i] == 0)
		{
			noOfZero = i;
			break;
		}
	}

	

	// sort for the first boundJ elements
/*	for (i = 0; i < boundJ-1; i++)
	{
		for (j = 0; j < boundJ-1-i; j++)
		{
			if (element[j] < element[j+1])
			{
				int temp;
				short temp2;

				// swap content
				temp = element[j+1];
				element[j+1] = element[j];
				element[j] = temp;

				// swap index
				temp2 = returnValue[j+1];
				returnValue[j+1] = returnValue[j];
				returnValue[j] = temp2;
			}
		}
	}
*/

	if (noOfZero >= boundJ)
	{
		// no need to sort (as all zeros in buondJ)
	}
	else if (noOfZero < boundJ)
	{
		quicksort_i_indShort(returnValue, element, noOfZero, boundJ-1);
	//}
	//else
	//{
		//void quicksort_i_indShort(short *sorted_index, int *value, int start_index, int end_index)
		//quicksort_i_indShort(returnValue, element, 0, boundJ-1);
	

		// check the remaining elts are the boundJ largest
		for (i = boundJ; i < noOfItem; i++)
		{
			if (element[i] < element[boundJ-1])
			{
				for (j = boundJ-1; j >= noOfZero; j--)
				{
					if (element[i] < element[j])
					{
						index = j;
						value = element[i];
						value2 = returnValue[i];
						//value2 = (short)i;
						break;
					}
				}

				// move content
				element[i] = element[boundJ-1];
				for (j = boundJ-1; j > index; j--)
				{
					element[j] = element[j-1];
				}
				element[index] = value;

				// move index
				//returnValue[i] = returnValue[boundJ-1];
				for (j = boundJ-1; j > index; j--)
				{
					returnValue[j] = returnValue[j-1];
				}
				returnValue[index] = value2;
			}
		}

		// print sorted index
		//printf("Sorted index:\n");
		//for (i = 0; i < boundJ; i++)
		//{
		//	printf("%d ", returnValue[i]);
		//}
		//printf("\n");
	}

	memcpy(returnValue2, returnValue, sizeof(short)*boundJ);

	free(returnValue);
	
	
	//(*noOfJ) = boundJ;
	noOfNonZero = noOfItem - noOfZero;

	//if (noOfNonZero < J)
	//{
	//	(*noOfJ) = noOfNonZero;
	//}
	//else
	//{
		(*noOfJ) = J;
	//}
	//for (i = 0; i < boundJ; i++)
	//{
	//	if (element[i] == 0)
	//	{
	//		(*noOfJ) = i+1;
	//		break;
	//	}
	//}
	
	return returnValue2;
}

/*
 * Function  : getConfOLD (dummy)
 * Purpose   : to calculate the confidence of the rule (defined previously, NOT currently used)
 *             (i.e. conf(d -> i))
 * Parameters: 
 *   1. data: the data in the database
 *   2. noOfTrans: no. of transactions
 *   3. d: the set of unselected items d
 *   4. noOfElt: no. of elements in d (i.e. |d|)
 *   5. i: item I in the loss rule
 * NOTE:
 *   1. the rule is
 *         not d -> not I_{i}
 *   2. Assume d does not contain I_i
 *   3. if noOfElt = 0 then the confidence = 1.0
 */
double getConfOLD(transPtr *data, int noOfTrans, int *d, int noOfElt, int i)
{
	double returnValue;
	double numerator, denumerator;

	int *tempDI;

	if (noOfElt == 0)
	{
		return 1.0;
	}

	tempDI = (int *) malloc(sizeof(int)*(noOfElt+1));
	for (i = 0; i < noOfElt; i++)
	{
		tempDI[i] = d[i];
	}
	tempDI[noOfElt] = i;

	//dI/d
	numerator = getSupport(data, noOfTrans, tempDI, noOfElt+1);
	denumerator = getSupport(data, noOfTrans, d, noOfElt);

	if (denumerator != 0)
	{
		returnValue = numerator/denumerator;
	}
	else
	{
		returnValue = 0;
	}

	free(tempDI);

	return returnValue;
}

/*
 * Function  : calculateBenefit
 * Purpose   : to calculate the benefit of item itemI
 * Parameters: 
 *   1. data: the data in the database
 *   2. noOfTrans: no. of transactions
 *   3. matrix: the matrix storing the support of the co-occurence
 *   4. elt: the estimation set S
 *   5. noOfElt: no. of elements
 *   6. itemI: item I
 *   7. profit: profit of item I
 */
double calculateBenefit(transPtr *data, int noOfTrans, short **matrix, int *elt, int noOfElt, int itemI, int profit)
{
	double returnValue = 0;

	int aItem;

	int *d;
	int noOfD;

	int i, j;//, k;

	int *isChosen;
	int tempNo;

	aItem = itemI;

	isChosen = (int *) malloc(sizeof(int)*noOfItem);
	memset(isChosen, 0, sizeof(int)*noOfItem);

	// initialize isChosen
	for (i = 0; i < noOfElt; i++)
	{
		isChosen[elt[i]] = 1;
	}


	d = (int *) malloc(sizeof(int)*maxTransSize);

	for (i = 0; i < noOfTrans; i++)
	{
		if (isContain(data[i], &aItem, 1) == 1)
		{
			// find d
			noOfD = 0;
			for (j = 0; j < data[i]->noOfItem; j++)
			{

				//isChosen = 0;
				//for (k = 0; k < noOfElt; k++)
				//{
				//	if (elt[k] == data[i]->item[j])
				//	{
				//		isChosen = 1;
				//		break;
				//	}
				//}

				tempNo = data[i]->item[j];

				if (isChosen[tempNo] == 0)
				{
					d[noOfD] = tempNo;
					noOfD++;
				}
			}
			//returnValue += profit*getConfOLD(data, noOfTrans, d, noOfD, itemI);
			returnValue += profit*(1-getConf(data, noOfTrans, matrix, d, noOfD, itemI));
		}
	}

	free(isChosen);
	free(d);

	return returnValue;
}

/*
 * Function  : calculateBenefit2
 * Purpose   : to calculate the benefit of item itemI
 * Parameters: 
 *   1. data: the data in the database
 *   2. noOfTrans: no. of transactions
 *   3. matrix: the matrix storing the support of the co-occurence
 *   4. elt: the estimation set S
 *   5. noOfElt: no. of elements
 *   6. itemI: item I
 *   7. profit: profit of item I
 */
double calculateBenefit2(transPtr *data, int noOfTrans, short **matrix, int *elt, int noOfElt, int *profit)
{
	double returnValue = 0;

	int itemI;

	int *d;
	int noOfD;

	int *sel;
	int noOfSel;

	int i, j;//, k;

	int *isChosen;
	int tempNo;

	//aItem = itemI;

	isChosen = (int *) malloc(sizeof(int)*noOfItem);
	memset(isChosen, 0, sizeof(int)*noOfItem);

	// initialize isChosen
	for (i = 0; i < noOfElt; i++)
	{
		isChosen[elt[i]] = 1;
	}


	d = (int *) malloc(sizeof(int)*maxTransSize);
	sel = (int *) malloc(sizeof(int)*maxTransSize);

	for (i = 0; i < noOfTrans; i++)
	{
		//if (isContain(data[i], &aItem, 1) == 1)
		if (isContainAny(data[i], isChosen) == 1)
		{
			// find d
			noOfD = 0;
			noOfSel = 0;
			for (j = 0; j < data[i]->noOfItem; j++)
			{

				//isChosen = 0;
				//for (k = 0; k < noOfElt; k++)
				//{
				//	if (elt[k] == data[i]->item[j])
				//	{
				//		isChosen = 1;
				//		break;
				//	}
				//}

				tempNo = data[i]->item[j];

				if (isChosen[tempNo] == 0)
				{
					d[noOfD] = tempNo;
					noOfD++;
				}
				else
				{
					sel[noOfSel] = tempNo;
					noOfSel++;
				}
			}

			if (noOfD == 0)
			{
				// all items are selected

				for (j = 0; j < noOfSel; j++)
				{
					itemI = sel[j];

					returnValue += profit[itemI];
				}
				
			}
			else if (noOfSel == 0)
			{
				// all items are not selected

				// do nothing
			}
			else
			{
				// some items are selected and some items are not selected

				for (j = 0; j < noOfSel; j++)
				{
					itemI = sel[j];
					//returnValue += profit*getConfOLD(data, noOfTrans, d, noOfD, itemI);
					returnValue += profit[itemI]*(1-getConf(data, noOfTrans, matrix, d, noOfD, itemI));
				}
			}
		}
	}

	free(isChosen);
	free(d);
	free(sel);

	return returnValue;
}



/*
 * Function  : calculateBenefit3
 * Purpose   : to calculate the benefit of item itemI
 * Parameters: 
 *   1. data: the data in the database
 *   2. noOfTrans: no. of transactions
 *   3. matrix: the matrix storing the support of the co-occurence
 *   4. elt: the estimation set S
 *   5. noOfElt: no. of elements
 *   6. itemI: item I
 *   7. profit: profit of item I
 */
double calculateBenefit3(transPtr *data, int noOfTrans, short **matrix, int *elt, int noOfElt, int *profit)
{
	double returnValue = 0;
	double returnValue2 = 0;
	double returnValue3 = 0;

	int itemI;

	int itemSpecial;

//	int *d;
//	int noOfD;

//	int *sel;
//	int noOfSel;

//	int *d2;
//	int noOfD2;

//	int *sel2;
//	int noOfSel2;

	int i, j;//, k;

	int *isChosen;
	int tempNo;

	double temp;

	int flag;

	//aItem = itemI;

	itemSpecial = elt[noOfElt-1];

	isChosen = (int *) malloc(sizeof(int)*noOfItem);
	memset(isChosen, 0, sizeof(int)*noOfItem);

	// initialize isChosen
	for (i = 0; i < noOfElt; i++)
	{
		isChosen[elt[i]] = 1;
	}


//	d = (int *) malloc(sizeof(int)*maxTransSize);
//	sel = (int *) malloc(sizeof(int)*maxTransSize);

//	d2 = (int *) malloc(sizeof(int)*maxTransSize);
//	sel2 = (int *) malloc(sizeof(int)*maxTransSize);


	for (i = 0; i < noOfTrans; i++)
	{
//if (i%1000 == 0) 
//printf("   t%d\n", i);
		flag = isContainAny2(data[i], isChosen, itemSpecial);

		//if (isContain(data[i], &aItem, 1) == 1)
		if (flag != 0)
		{
			// find d
			noOfD = 0;
			noOfSel = 0;
			for (j = 0; j < data[i]->noOfItem; j++)
			{

				//isChosen = 0;
				//for (k = 0; k < noOfElt; k++)
				//{
				//	if (elt[k] == data[i]->item[j])
				//	{
				//		isChosen = 1;
				//		break;
				//	}
				//}

				tempNo = data[i]->item[j];

				if (isChosen[tempNo] == 0)
				{
					d[noOfD] = tempNo;
					noOfD++;
				}
				else
				{
					sel[noOfSel] = tempNo;
					noOfSel++;
				}
			}


			// calculate returnValue
			temp = 0;
			if (noOfD == 0)
			{
				// all items are selected

				for (j = 0; j < noOfSel; j++)
				{
					itemI = sel[j];

					temp += profit[itemI];
				}
				returnValue += temp;
			}
			else if (noOfSel == 0)
			{
				// all items are not selected

				// do nothing
			}
			else
			{
				// some items are selected and some items are not selected
				for (j = 0; j < noOfSel; j++)
				{
					itemI = sel[j];
					//returnValue += profit*getConfOLD(data, noOfTrans, d, noOfD, itemI);
					temp += profit[itemI]*(1-getConf(data, noOfTrans, matrix, d, noOfD, itemI));
				}
				returnValue += temp;
			}


			
			if (flag == 1)
			{
				returnValue2 += temp;
			}
			else if (flag == 2)
			{

				if (noOfSel == 0)
				{
					// do nothing
				}
				else
				{
					// some items are selected and some items are not selected

					d2[noOfD] = itemSpecial;
					noOfD2 = noOfD + 1;

					noOfSel2 = 0;
					for (j = 0; j < noOfSel; j++)
					{
						if (sel[j] != itemSpecial)
						{
							sel2[noOfSel2] = sel[j];
							noOfSel2++;
						}
					}
				
					for (j = 0; j < noOfSel2; j++)
					{
						itemI = sel2[j];
						returnValue2 += profit[itemI]*(1-getConf(data, noOfTrans, matrix, d2, noOfD2, itemI));
					}
				}
			}

		}
	}

	free(isChosen);
	//free(d);
	//free(sel);
	//free(d2);
	//free(sel2);

	returnValue3 = returnValue - returnValue2;

	return returnValue3;
}











/*
 * Function  : calculateBenefitFP
 * Purpose   : to calculate the benefit of item itemI by using FP-tree approach
 * Parameters: 
 *   1. data: the data in the database
 *   2. noOfTrans: no. of transactions
 *   3. matrix: the matrix storing the support of the co-occurence
 *   4. elt: the estimation set S
 *   5. noOfElt: no. of elements
 *   6. itemI: item I
 *   7. profit: profit of item I
 */
double calculateBenefitFP(transPtr *data, int noOfTrans, short **matrix, int *elt, int noOfElt, int *profit, double *revNo, int *chooseFlag)
{
	double returnValue = 0;

	//int itemI;

	//int *d;
	//int noOfD;

	//int *sel;
	//int noOfSel;

	int noOfNonSel;

	int i, j;//, k;

	int *isChosen;
	//int tempNo;

	int *itemList;
	int *newSupport;

	short *fpMap;

	int startIndex, endIndex;

	//aItem = itemI;

	isChosen = (int *) malloc(sizeof(int)*noOfItem);
	//memset(isChosen, 0, sizeof(int)*noOfItem);

	itemList = (int *) malloc(sizeof(int)*noOfItem);

	newSupport = (int *) malloc(sizeof(int)*noOfItem);
	//memset(newSupport, 0, sizeof(int)*noOfItem);

	fpMap = (short *) malloc(sizeof(short)*noOfItem);
	//memset(fpMap, 0, sizeof(short)*noOfItem);

	if ((isChosen == NULL) || (itemList == NULL) || (newSupport == NULL) || (fpMap == NULL))
	{
		printf("Insufficent memory!\n");
		exit(1);
	}
	memset(isChosen, 0, sizeof(int)*noOfItem);
	memset(newSupport, 0, sizeof(int)*noOfItem);
	memset(fpMap, 0, sizeof(short)*noOfItem);

	// initialize isChosen
	for (i = 0; i < noOfElt; i++)
	{
		isChosen[elt[i]] = 1;
		
	}

	// initialize itemList and newSupport
	for (i = 0; i < noOfItem; i++)
	{
		itemList[i] = i;
		newSupport[i] = support1[i];
	}


	// Make Selected Items and Unselected items separate

	startIndex = -1;
	endIndex = noOfItem;
	while (1)
	{
		do
		{
			startIndex++;
		}
		while ((startIndex < noOfItem) && (isChosen[itemList[startIndex]] == 0));

		do
		{
			endIndex--;
		}
		while ((endIndex >= 0) && (isChosen[itemList[endIndex]] == 1));

		// swap
		if (startIndex < endIndex)
		{
			//void q_swap_i(int *sorted_index, int *value, int i, int j)
			q_swap_i(itemList, newSupport, startIndex, endIndex);
		}
		else
		{
			break;
		}
	}

	// here, all unselected items are before selected items in itemList

	noOfNonSel = noOfItem - J;
	
	// Now, we need to sort unselected and selected items separately according to newSupport
	//void quicksort(int *sorted_index, double *value, int start_index, int end_index)
	//quicksort_i(returnValue, element, 0, noOfNonZero-1);
	quicksort_i(itemList, newSupport, 0, noOfNonSel-1);

	quicksort_i(itemList, newSupport, noOfNonSel, noOfItem-1);
	

	// create a FP-map
	for (i = 0; i < noOfItem; i++)
	{
		for (j = 0; j < noOfItem; j++)
		{
			if (j == itemList[i])
			{
				fpMap[j] = (short) i;
				break;
			}
		}
	}	



	// call buildTree

	// calculate benefit in FP-tree 
	//double buildTreeSet(transPtr *data, int *fpMap, int *isChoose, int *profit, double *revNo)
	returnValue = buildTreeSet(data, fpMap, isChosen, profit, revNo, chooseFlag);


	// free memory
	free(isChosen);
	free(itemList);
	free(newSupport);
	free(fpMap);
	


/*

	d = (int *) malloc(sizeof(int)*maxTransSize);
	sel = (int *) malloc(sizeof(int)*maxTransSize);

	for (i = 0; i < noOfTrans; i++)
	{
		//if (isContain(data[i], &aItem, 1) == 1)
		if (isContainAny(data[i], isChosen) == 1)
		{
			// find d
			noOfD = 0;
			noOfSel = 0;
			for (j = 0; j < data[i]->noOfItem; j++)
			{

				//isChosen = 0;
				//for (k = 0; k < noOfElt; k++)
				//{
				//	if (elt[k] == data[i]->item[j])
				//	{
				//		isChosen = 1;
				//		break;
				//	}
				//}

				tempNo = data[i]->item[j];

				if (isChosen[tempNo] == 0)
				{
					d[noOfD] = tempNo;
					noOfD++;
				}
				else
				{
					sel[noOfSel] = tempNo;
					noOfSel++;
				}
			}

			if (noOfD == 0)
			{
				// all items are selected

				for (j = 0; j < noOfSel; j++)
				{
					itemI = sel[j];

					returnValue += profit[itemI];
				}
				
			}
			else if (noOfSel == 0)
			{
				// all items are not selected

				// do nothing
			}
			else
			{
				// some items are selected and some items are not selected

				for (j = 0; j < noOfSel; j++)
				{
					itemI = sel[j];
					//returnValue += profit*getConfOLD(data, noOfTrans, d, noOfD, itemI);
					returnValue += profit[itemI]*(1-getConf(data, noOfTrans, matrix, d, noOfD, itemI));
				}
			}
		}
	}

	free(isChosen);
	free(d);
	free(sel);
*/
	//if (noOfElt > noOfItem*0.5)
	//{
	//	returnValue += profit[elt[noOfElt-1]];
	//}

	return returnValue;
}







/*
 * Function  : calculateBenefitFP2
 * Purpose   : to calculate the benefit of item itemI by using FP-tree approach
 * Parameters: 
 *   1. data: the data in the database
 *   2. noOfTrans: no. of transactions
 *   3. matrix: the matrix storing the support of the co-occurence
 *   4. elt: the estimation set S
 *   5. noOfElt: no. of elements
 *   6. itemI: item I
 *   7. profit: profit of item I
 */
double calculateBenefitFP2(transPtr *data, int noOfTrans, short **matrix, int *elt, int noOfElt, int *profit, double *revNo, int *chooseFlag, int itemSpecial)
{
	double returnValue = 0;

	//int itemI;

	//int *d;
	//int noOfD;

	//int *sel;
	//int noOfSel;

	int noOfNonSel;

	int i, j;//, k;

	int *isChosen;
	//int tempNo;

	int *itemList;
	int *newSupport;

	short *fpMap;

	int startIndex, endIndex;

	//aItem = itemI;

	isChosen = (int *) malloc(sizeof(int)*noOfItem);
	//memset(isChosen, 0, sizeof(int)*noOfItem);

	itemList = (int *) malloc(sizeof(int)*noOfItem);

	newSupport = (int *) malloc(sizeof(int)*noOfItem);
	//memset(newSupport, 0, sizeof(int)*noOfItem);

	fpMap = (short *) malloc(sizeof(short)*noOfItem);
	//memset(fpMap, 0, sizeof(short)*noOfItem);

	if ((isChosen == NULL) || (itemList == NULL) || (newSupport == NULL) || (fpMap == NULL))
	{
		printf("Insufficent memory!\n");
		exit(1);
	}
	memset(isChosen, 0, sizeof(int)*noOfItem);
	memset(newSupport, 0, sizeof(int)*noOfItem);
	memset(fpMap, 0, sizeof(short)*noOfItem);

	// initialize isChosen
	for (i = 0; i < noOfElt; i++)
	{
		isChosen[elt[i]] = 1;
		
	}

	// initialize itemList and newSupport
	for (i = 0; i < noOfItem; i++)
	{
		itemList[i] = i;
		newSupport[i] = support1[i];
	}


	// Make Selected Items and Unselected items separate

	startIndex = -1;
	endIndex = noOfItem;
	while (1)
	{
		do
		{
			startIndex++;
		}
		while ((startIndex < noOfItem) && (isChosen[itemList[startIndex]] == 0));

		do
		{
			endIndex--;
		}
		while ((endIndex >= 0) && (isChosen[itemList[endIndex]] == 1));

		// swap
		if (startIndex < endIndex)
		{
			//void q_swap_i(int *sorted_index, int *value, int i, int j)
			q_swap_i(itemList, newSupport, startIndex, endIndex);
		}
		else
		{
			break;
		}
	}

	// here, all unselected items are before selected items in itemList


	// make the specialItem in the boundary

	noOfNonSel = noOfItem - J;
	for (i = 0; i < noOfItem; i++)
	{
		if (itemList[i] == itemSpecial)
		{
			// swap with the elt at the boundary between selected and unselected
			q_swap_i(itemList, newSupport, i, noOfNonSel);
			break;
		}
	}
	
	// Now, we need to sort unselected and selected items separately according to newSupport
	//void quicksort(int *sorted_index, double *value, int start_index, int end_index)
	//quicksort_i(returnValue, element, 0, noOfNonZero-1);
	quicksort_i(itemList, newSupport, 0, noOfNonSel-1);

	quicksort_i(itemList, newSupport, noOfNonSel+1, noOfItem-1);
	

	// create a FP-map
	for (i = 0; i < noOfItem; i++)
	{
		for (j = 0; j < noOfItem; j++)
		{
			if (j == itemList[i])
			{
				fpMap[j] = (short) i;
				break;
			}
		}
	}	



	// call buildTree

	// calculate benefit in FP-tree 
	//double buildTreeSet2(transPtr *data, short *fpMap, int *isChoose, int *profit, double *revNo, int *chooseFlag, int itemSpecial)
	returnValue = buildTreeSet2(data, fpMap, isChosen, profit, revNo, chooseFlag, itemSpecial);



	// free memory
	free(isChosen);
	free(itemList);
	free(newSupport);
	free(fpMap);
	


	return returnValue;
}










/*
 * Function  : calculateBenefitFP_diff
 * Purpose   : to calculate the benefit difference of item itemI by using FP-tree approach
 *             before the seleciton and after the selection
 * Parameters: 
 *   1. data: the data in the database
 *   2. noOfTrans: no. of transactions
 *   3. matrix: the matrix storing the support of the co-occurence
 *   4. elt: the estimation set S
 *   5. noOfElt: no. of elements
 *   6. itemI: item I
 *   7. profit: profit of item I
 */
double calculateBenefitFP_diff(transPtr *data, int noOfTrans, short **matrix, int *elt, int noOfElt, int *profit, double *revNo, int *chooseFlag, int itemAdd, int itemRemove)
{
	double returnValue = 0;

	//int itemI;

	//int *d;
	//int noOfD;

	//int *sel;
	//int noOfSel;

	int noOfNonSel;

	int i, j;//, k;

	int *isChosen;
	//int tempNo;

	int *itemList;
	int *newSupport;

	short *fpMap;
	short *fpMap2;

	int itemNo;
	int temp;

	int startIndex, endIndex;

	//aItem = itemI;

	isChosen = (int *) malloc(sizeof(int)*noOfItem);
	//memset(isChosen, 0, sizeof(int)*noOfItem);

	itemList = (int *) malloc(sizeof(int)*noOfItem);

	newSupport = (int *) malloc(sizeof(int)*noOfItem);
	//memset(newSupport, 0, sizeof(int)*noOfItem);

	fpMap = (short *) malloc(sizeof(short)*noOfItem);
	fpMap2 = (short *) malloc(sizeof(short)*noOfItem);
	//memset(fpMap, 0, sizeof(short)*noOfItem);
	//memset(fpMap2, 0, sizeof(short)*noOfItem);

	if ((isChosen == NULL) || (itemList == NULL) || (newSupport == NULL) || (fpMap == NULL) || (fpMap2 == NULL))
	{
		printf("Insufficent memory!\n");
		exit(1);
	}
	memset(isChosen, 0, sizeof(int)*noOfItem);
	memset(newSupport, 0, sizeof(int)*noOfItem);
	memset(fpMap, 0, sizeof(short)*noOfItem);
	memset(fpMap2, 0, sizeof(short)*noOfItem);

	// initialize isChosen
	for (i = 0; i < noOfElt; i++)
	{
		isChosen[elt[i]] = 1;
		
	}

	// initialize itemList and newSupport
	for (i = 0; i < noOfItem; i++)
	{
		itemList[i] = i;
		newSupport[i] = support1[i];
	}


	// Make Selected Items and Unselected items separate

	startIndex = -1;
	endIndex = noOfItem;
	while (1)
	{
		do
		{
			startIndex++;
		}
		while ((startIndex < noOfItem) && (isChosen[itemList[startIndex]] == 0));

		do
		{
			endIndex--;
		}
		while ((endIndex >= 0) && (isChosen[itemList[endIndex]] == 1));

		// swap
		if (startIndex < endIndex)
		{
			//void q_swap_i(int *sorted_index, int *value, int i, int j)
			q_swap_i(itemList, newSupport, startIndex, endIndex);
		}
		else
		{
			break;
		}
	}

	// here, all unselected items are before selected items in itemList

	noOfNonSel = noOfItem - J;
	

	for (i = 0; i < noOfItem; i++)
	{
		itemNo = itemList[i];

		if (itemNo == itemAdd)
		{
			q_swap_i(itemList, newSupport, i, noOfNonSel);
		}
		else if (itemNo == itemRemove)
		{
			q_swap_i(itemList, newSupport, i, noOfNonSel-1);
		}
	}

	// Now, we need to sort unselected and selected items separately according to newSupport
	//void quicksort(int *sorted_index, double *value, int start_index, int end_index)
	//quicksort_i(returnValue, element, 0, noOfNonZero-1);
	quicksort_i(itemList, newSupport, 0, noOfNonSel-1);

	quicksort_i(itemList, newSupport, noOfNonSel, noOfItem-1);
	

	// create a FP-map
	for (i = 0; i < noOfItem; i++)
	{
		for (j = 0; j < noOfItem; j++)
		{
			if (j == itemList[i])
			{
				fpMap[j] = (short) i;
				break;
			}
		}
	}	

	// create a FP-map2
	memcpy(fpMap2, fpMap, sizeof(short)*noOfItem);

	temp = fpMap2[itemAdd];
	fpMap2[itemAdd] = fpMap2[itemRemove];
	fpMap2[itemRemove] = temp;


	// call buildTree

	// calculate benefit in FP-tree 
	//double buildTreeSet(transPtr *data, int *fpMap, int *isChoose, int *profit, double *revNo)
	returnValue = buildTreeSet_diff(data, fpMap, fpMap2, isChosen, profit, revNo, chooseFlag, itemAdd, itemRemove);


	// free memory
	free(isChosen);
	free(itemList);
	free(newSupport);
	free(fpMap);
	free(fpMap2);
	


	return returnValue;
}





/*
 * Function  : transform2Matrix
 * Purpose   : to transfrom the freq. itemsets to a matrix form
 * Parameters: 
 *   1. freqitemset: freq. itemsets
 *   2. noOfElt: no. of freq. itemsets
 * NOTE:
 *   1. the freq. itemsets of size 2 should be sorted first
 */
short **transform2Matrix(largeitemPtr *freqitemset, int noOfElt)
{
	short **returnValue = NULL;
	int i, j;

	int hashNo;

	int itemI, itemJ;
	int setFreq;

	int no;
	short **tempElt;
	int *tempFreq;
	int temp;

	//returnValue = (short **) malloc(sizeof(short *)*noOfElt);
	//for (i = 0; i < noOfElt; i++)
	//{
	//	returnValue[i] = (short *) malloc(sizeof(short)*noOfElt);
	//	for (j = 0; j < noOfElt; j++)
	//	{
	//		returnValue[i][j] = (short) 0;
	//	}
	//}

	// for different items of size two
	//for (i = 0; i < freqitemset[1]->noOfItemSet; i++)
	//{
	//	returnValue[freqitemset[1]->item[i]->itemNo[0]][freqitemset[1]->item[i]->itemNo[1]] += (short) freqitemset[1]->item[i]->freq;
	//}

	// for same item of size one
	//for (i = 0; i < freqitemset[0]->noOfItemSet; i++)
	//{
	//	int no;
	//	no = freqitemset[0]->item[i]->itemNo[0];
	//	returnValue[no][no] = (short) freqitemset[0]->item[i]->freq;
	//}

#ifdef MATRIX
	printf("------------\n");
	matrix = (short **) malloc(sizeof(short *)*noOfItem);
	for (i = 0; i < noOfItem; i++)
	{
		matrix[i] = (short *) malloc(sizeof(short)*noOfItem);
		memset(matrix[i], 0, sizeof(short)*noOfItem);
	}
#endif

	// for different items of size two 
	for (i = 0; i < freqitemset[1]->noOfItemSet; i++)
	{


		itemI = freqitemset[1]->item[i]->itemNo[0];
		itemJ = freqitemset[1]->item[i]->itemNo[1];
		setFreq = freqitemset[1]->item[i]->freq;

		if (itemI > itemJ)
		{
			//int temp;

			temp = itemI;
			itemI = itemJ;
			itemJ = temp;
		}

#ifdef MATRIX
		matrix[itemI][itemJ] = setFreq;
#else

		hashNo = (itemI*H_START + itemJ)%NO_SLOT;

		if (hashMatrix[hashNo].noOfElt > 0)
		{


			no = hashMatrix[hashNo].noOfElt;

			// copy
			tempElt = (short **) malloc(sizeof(short *)*no);
			tempFreq = (int *) malloc(sizeof(int)*no);
			for (j = 0; j < no; j++)
			{
				tempElt[j] = (short *) malloc(sizeof(short)*2);
				
				tempElt[j][0] = hashMatrix[hashNo].elt[j][0];
				tempElt[j][1] = hashMatrix[hashNo].elt[j][1];

				free(hashMatrix[hashNo].elt[j]);

				tempFreq[j] = hashMatrix[hashNo].freq[j];
			}
			free(hashMatrix[hashNo].elt);
			free(hashMatrix[hashNo].freq);

			// create and copy back
			hashMatrix[hashNo].elt = (short **) malloc(sizeof(short *)*(no+1));
			hashMatrix[hashNo].freq = (int *) malloc(sizeof(int)*(no+1));
			for (j = 0; j < no; j++)
			{
				hashMatrix[hashNo].elt[j] = (short *) malloc(sizeof(short)*2);

				hashMatrix[hashNo].elt[j][0] = tempElt[j][0];
				hashMatrix[hashNo].elt[j][1] = tempElt[j][1];

				free(tempElt[j]);

				hashMatrix[hashNo].freq[j] = tempFreq[j];
			}
			free(tempElt);
			free(tempFreq);


			hashMatrix[hashNo].elt[no] = (short *) malloc(sizeof(short)*2);

			//add new elt
			hashMatrix[hashNo].elt[no][0] = (short) itemI;
			hashMatrix[hashNo].elt[no][1] = (short) itemJ;
			hashMatrix[hashNo].freq[no] = setFreq;

			hashMatrix[hashNo].noOfElt = no+1;
		}
		else
		{
			// only 1 elemtn
			hashMatrix[hashNo].elt = (short **) malloc(sizeof(short *));
			hashMatrix[hashNo].freq = (int *) malloc(sizeof(int));
			hashMatrix[hashNo].elt[0] = (short *) malloc(sizeof(short)*2);

			// add new elt
			hashMatrix[hashNo].elt[0][0] = itemI;
			hashMatrix[hashNo].elt[0][1] = itemJ;
			hashMatrix[hashNo].freq[0] = setFreq;

			hashMatrix[hashNo].noOfElt = 1;
		}

#endif

	}

	
	// for the same size of one
	for (i = 0; i < freqitemset[0]->noOfItemSet; i++)
	{
		//int hashNo;

		//int itemI, itemJ;
		//int setFreq;

		itemI = freqitemset[0]->item[i]->itemNo[0];
		//itemJ = freqitemset[0]->item[i]->itemNo[1];
		itemJ = itemI;
		setFreq = freqitemset[0]->item[i]->freq;

		if (itemI > itemJ)
		{
			

			temp = itemI;
			itemI = itemJ;
			itemJ = temp;
		}

#ifdef MATRIX
		matrix[itemI][itemJ] = setFreq;
#else

		hashNo = (itemI*H_START + itemJ)%NO_SLOT;

		if (hashMatrix[hashNo].noOfElt > 0)
		{
			//int no;

			//short **tempElt;
			//int *tempFreq;

			no = hashMatrix[hashNo].noOfElt;

			// copy
			tempElt = (short **) malloc(sizeof(short *)*no);
			tempFreq = (int *) malloc(sizeof(int)*no);
			for (j = 0; j < no; j++)
			{
				tempElt[j] = (short *) malloc(sizeof(short)*2);
				
				tempElt[j][0] = hashMatrix[hashNo].elt[j][0];
				tempElt[j][1] = hashMatrix[hashNo].elt[j][1];

				free(hashMatrix[hashNo].elt[j]);

				tempFreq[j] = hashMatrix[hashNo].freq[j];
			}
			free(hashMatrix[hashNo].elt);
			free(hashMatrix[hashNo].freq);

			// create and copy back
			hashMatrix[hashNo].elt = (short **) malloc(sizeof(short *)*(no+1));
			hashMatrix[hashNo].freq = (int *) malloc(sizeof(int)*(no+1));
			for (j = 0; j < no; j++)
			{
				hashMatrix[hashNo].elt[j] = (short *) malloc(sizeof(short)*2);

				hashMatrix[hashNo].elt[j][0] = tempElt[j][0];
				hashMatrix[hashNo].elt[j][1] = tempElt[j][1];

				free(tempElt[j]);

				hashMatrix[hashNo].freq[j] = tempFreq[j];
			}
			free(tempElt);
			free(tempFreq);


			hashMatrix[hashNo].elt[no] = (short *) malloc(sizeof(short)*2);

			//add new elt
			hashMatrix[hashNo].elt[no][0] = (short) itemI;
			hashMatrix[hashNo].elt[no][1] = (short) itemJ;
			hashMatrix[hashNo].freq[no] = setFreq;

			hashMatrix[hashNo].noOfElt = no+1;
		}
		else
		{
			// only 1 elemtn
			hashMatrix[hashNo].elt = (short **) malloc(sizeof(short *));
			hashMatrix[hashNo].freq = (int *) malloc(sizeof(int));
			hashMatrix[hashNo].elt[0] = (short *) malloc(sizeof(short)*2);

			// add new elt
			hashMatrix[hashNo].elt[0][0] = itemI;
			hashMatrix[hashNo].elt[0][1] = itemJ;
			hashMatrix[hashNo].freq[0] = setFreq;

			hashMatrix[hashNo].noOfElt = 1;
		}
#endif
	}



	



	return returnValue;
}

/*
// freq of size 1
int *transform2Vector(largeitemPtr *freqitemset, int noOfElt)
{
	int *returnValue;

	returnValue = (int *) malloc(sizeof(int)*noOfElt);
	for (i = 0; i < noOfElt; i++)
	{
		returnValue[i] = 0;
	}

	for (i = 0; i < freqitemset[0]->noOfItemSet; i++)
	{
		returnValue[freqitemset[0]->item[i]->itemNo[0]] = freqitemset[0]->item[i]->freq;
	}

	return returnValue;
}
*/

/*
 * Function  : calculateUnion
 * Purpose   : to calculate the transactions containing the elements in the set
 * Parameters: 
 *   1. matrix: the storage storing the co-occurences and size-1 itemsets
 *   2. elt: the set of elements
 *   3. noOfElt: no. of elements
 */
int calculateUnion(short **matrix, int *elt, int noOfElt)
{
	int i, j;
	int returnValue = 0;

	//Sorting the elt (in ascending)
	for (i = 0; i < noOfElt-1; i++)
	{
		for (j = 0; j < noOfElt-1-i; j++)
		{
			if (elt[j] > elt[j+1])
			{
				int temp;

				//swapping
				temp = elt[j];
				elt[j] = elt[j+1];
				elt[j+1] = temp;
			}
		}
	}
	
	// for size 1
	for (i = 0; i < noOfElt; i++)
	{
		int no;
		no = elt[i];
		//returnValue += matrix[no][no];
		returnValue += getMatrix(no,no);
	}

	// for size 2
	for (i = 0; i < noOfElt; i++)
	{
		for (j = i+1; j < noOfElt; j++)
		{
			int no1 = elt[i];
			int no2 = elt[j];

			//returnValue -= matrix[no1][no2];
			returnValue -= getMatrix(no1,no2);
		}
	}

	if (returnValue < 0)
	{
		returnValue = 0;
	}

	return returnValue;
}


/*
 * Function  : calculateUnionTwo
 * Purpose   : to calculate the union of no of elt which co-occur with Item I
 * Parameters: 
 *   1. data: the data in the database
 *   2. noOfTrans: no. of trans
 *   3. matrix: the matrix storing the co-occurences
 *   4. itemI: item I
 *   5. elt: the set of elements
 *   6. noOfElt: no. of elements
 * NOTE:
 *   1. Assume itemI is not in elt
 */
int calculateUnionTwo(transPtr *data, int noOfTrans, short **matrix, int itemI, int *elt, int noOfElt)
{
	int returnValue = 0;

	int i, j;
	
	//Sorting the elt (in ascending)
	for (i = 0; i < noOfElt-1; i++)
	{
		for (j = 0; j < noOfElt-1-i; j++)
		{
			if (elt[j] > elt[j+1])
			{
				int temp;

				//swapping
				temp = elt[j];
				elt[j] = elt[j+1];
				elt[j+1] = temp;
			}
		}
	}

	// for size 2
	for (i = 0; i < noOfElt; i++)
	{
		int no1 = itemI;
		int no2 = elt[i];

		// make no1 < no2
		if (no1 > no2)
		{
			int temp;
			
			//swapping
			temp = no1;
			no1 = no2;
			no2 = temp;
		}

		//returnValue += matrix[no1][no2];
		returnValue += getMatrix(no1,no2);
	}

/*	// for size 3
	//** need to minus
	for (i = 0; i < noOfElt; i++)
	{
		for (j = i+1; j < noOfElt; j++)
		{
			int no[3];

			// make no in ascending order
			if (itemI < elt[i])
			{
				no[0] = itemI;
				no[1] = elt[i];
				no[2] = elt[j];
			}
			else if (itemI < elt[j])
			{
				no[0] = elt[i];
				no[1] = itemI;
				no[2] = elt[j];
			}
			else
			{
				no[0] = elt[i];
				no[1] = elt[j];
				no[2] = itemI;
			}

			returnValue -= getSupport(data, noOfTrans, no, 3);
		}
	}
*/

	// union is made to be non-negative
	if (returnValue < 0)
	{
		returnValue = 0;
	}

	return returnValue;
}

/*
 * Function  : getConf
 * Purpose   : to calculate the confidence of the loss rule 
 *             (i.e.i -> (elts or ..) where elt is in d
 * Parameters: 
 *   1. data: data in the database
 *   2. noOfTrans: no. of transactions
 *   3. matrix: the matrix storing the co-occurences
 *   4. d: the set of elements
 *   5. noOfElt: the number of elements in set d
 *   6. itemI: item I
 */
double getConf(transPtr *data, int noOfTrans, short **matrix, int *d, int noOfElt, int itemI)
{
	double numerator, denumerator;	

	double returnValue = 0.0;

	//numerator = calculateUnionTwo(data, noOfTrans, matrix, itemI, d, noOfElt);
	//denumerator = matrix[itemI][itemI];
	denumerator = getMatrix(itemI,itemI);

	numerator = getCountItemIAny(itemI, d, noOfElt);

	if (denumerator != 0)
	{

		// Make Sure the denumerator is larger
		if (numerator <= denumerator)
		{
			returnValue = numerator/denumerator;
		}
		else
		{
			returnValue = 1.0;
		}
	}
	else
	{
		returnValue = 0;
	}

	return returnValue;
}



/*
 * Function  : getMatrix
 * Purpose   : to get the entry in the matrix (itemI, itemJ)
 * Parameters: 
 *   1. itemI, itemJ: the parameters in the matrix
 * NOTE:
 *   1. Assume:   itemI < itemJ
 */

int getMatrix(int itemI, int itemJ)
{
	int returnValue;

	int hashNo;

	int isFound;

	int i;

	if (itemI > itemJ)
	{
		int temp;

		temp = itemI;
		itemI = itemJ;
		itemJ = temp;
	}

#ifdef MATRIX
	returnValue = matrix[itemI][itemJ];
	return returnValue;
#else

	hashNo = (itemI*H_START + itemJ)%NO_SLOT;

	isFound = 0;
	for (i = 0; i < hashMatrix[hashNo].noOfElt; i++)
	{
		if ((hashMatrix[hashNo].elt[i][0] == itemI) && (hashMatrix[hashNo].elt[i][1] == itemJ))
		{
			isFound = 1;
			returnValue = hashMatrix[hashNo].freq[i];
		}
	}

	if (isFound == 0)
	{
		returnValue = 0;
	}

	return returnValue;
#endif
}


/*
 * Function  : transform2Select
 * Purpose   : to transform the item-no representation to binary representation
 * Parameters: 
 *   1. s: the item-no representation
 *   2. S: the item-no representation
 *   3. J: no. of elements in S
 *   4. sortEstIndex2: sorted estimated index
 *   5. itemI: item I
 */
void transform2Select(int *s, short *S, int noOfJ, short *sortEstIndex2, int itemI)
{
	int ind = 0;

	int i, k;

	int temp;
	int *tempList = (int *) malloc(sizeof(int)*noOfItem);
//printf("---noOfJ (%d) : %d\n", itemI, noOfJ);
	for (i = 0; i < noOfItem; i++)
	{
		tempList[i] = 1;
	}


	for (i = 0; i < noOfItem-noOfJ; i++)
	{
		tempList[S[i]] = 0;
	}
	
	if (tempList[itemI] == 0)
	{
		printf("********** itemI %d is selected!\n", itemI);
	}
		

	for (k = 0; k < noOfItem; k++)
	{
		if (tempList[k] == 1)
		{
			if (k == itemI)
			{
				s[noOfJ-1] = k;
			}
			else
			{
				if (ind != noOfJ-1)
				{
					s[ind] = k;
					ind++;
				}
				
			}
		}
	}

	free(tempList);
}


