/*
 * Program        : readFP.h
 *
 * Objectives     : To read the fp-tree generated by FP-growth function
 *
 * Assumptions    : Item ID in dataset file should start from zero.
 *                  I.e. if there are NUM items in the dataset, 
 *                  the item ID should be from 0 to NUM-1.
 *
 * Variations     : 
 *
 * Platform       : SunOS/Linux
 * Compiler       : gcc
 *
 * Programmer     : Raymond Chi-Wing Wong
 * Creation Date  : Oct, 2002
 * Last Update    : Jun 9, 2004
 *
 * Reference      :
 *   1. J. Han, J. Pei, and Y. Yin, 
 *      "Mining Frequent Patterns without Candidate Generation", 
 *      Proc. ACM SIGMOD 2000.
 *   2. R. C.-W. Wong, A. W.-C. Fu and K. Wang
 *      "MPIS: Maximal-Profit Item Selection with Cross-Selling Considerations"
 *      The 2003 IEEE International Conference on Data Mining (ICDM)
 *      
 */

#ifndef READFP_H
#define READFP_H

#include <stdio.h>
#include <stdlib.h>

#include <string.h>

#define MAX_LEN 1024

typedef char *STRING;


struct itemset_s
{
	int *itemNo;
	int freq;
};

typedef struct itemset_s itemset;
typedef itemset *itemsetPtr;


struct largeitem_s
{
	int noOfItemSet;
	itemsetPtr *item;
};

typedef struct largeitem_s largeitem;
typedef largeitem *largeitemPtr;



void printUsage(char **argv);

largeitemPtr *readFP(STRING infile, int *upperLimit);

#endif

